// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the methods that are defined in the PBS_JobArray.h
//	header file.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include "PBS_JobArray.h"
#include "PBS_JobRecord.h"

// *****************************************************************************
// * PBS_JobArray::defaultSorter  :
// *	This is the default sort method that is used to sort PBS_JobRecords that 
// *	are added to the PBS_JobArray. By default the PBS_JobRecords will be sorted
// *	using the following parameters.
// *
// *	1) Finish Time		(ascending)
// *	2) Enqueued Time	(ascending)
// *	3) Start Time		(ascending)
// *	4) JobID		(ascending)
// *****************************************************************************
int PBS_JobArray::defaultSorter ( const PBS_JobRecord ** A, const PBS_JobRecord ** B)
	{
	const PBS_JobRecord *a = *A;
	const PBS_JobRecord *b = *B;
	
	int result;

	if     (a == b)  result = 0;
	else if(a && !b) result = 1;
	else if(!a && b) result = -1;
	else if(a->getTimeFinished() == b->getTimeFinished())
		{
		if(a->getTimeEnqueued() == b->getTimeEnqueued())
			{
			if(a->getTimeStarted() == b->getTimeStarted())
				{
		  		result = strcmp(a->getJobID(), b->getJobID());
				}
			else	{
				if(!a->getTimeStarted()) result = 1;
				else if(!b->getTimeStarted()) result = -1;
				else result = (int)(a->getTimeStarted()-b->getTimeStarted());
				}		
			}
		else	{
			if(!a->getTimeEnqueued()) result = 1;
			else if(!b->getTimeEnqueued()) result = -1;
			else result = (int)(a->getTimeEnqueued()-b->getTimeEnqueued());
			}
		}
	else	{
		if(!a->getTimeFinished()) result = 1;
		else if(!b->getTimeFinished()) result = -1;
		else result = (int)(a->getTimeFinished()-b->getTimeFinished());
		}
	return result;
	}
	

// *****************************************************************************
// * PBS_JobArray::PBS_JobArray  :
// *	The constructor for the class. It initializes the underlying and
// *	embedded objects and preallocates a specified number of PBS_JobRecord
// *	pointers.
// *****************************************************************************
PBS_JobArray::PBS_JobArray (const char * Name, size_t ArraySize)
	: PBS_JobCollection(Name, ArraySize),
	  records      (NULL),
	  arraySize    ((ArraySize<=0)?1:ArraySize),
	  recordCnt    (0)
	{
	records = (PBS_JobRecord **)malloc(arraySize*sizeof(PBS_JobRecord *));
	memset(records, 0, arraySize*sizeof(PBS_JobRecord *));
	}

// *****************************************************************************
// * PBS_JobArray::~PBS_JobArray  :
// *	The descructor.  It frees the PBS_JobRecord pointers that were allocated
// *	in the constructor or reallocated during the addition of new records.
// *****************************************************************************
PBS_JobArray::~PBS_JobArray ( void )
	{
	if(records) free(records);
	}
	
// *****************************************************************************
// * PBS_JobArray::clear  :
// *	Removes all records from the PBS_JobArray.  This method should be called 
// *	prior to reinitializing the object from the Server Logs.
// *****************************************************************************
void PBS_JobArray::clear ( void )
	{
	PBS_JobCollection::clear();
	memset(records, 0, arraySize*sizeof(PBS_JobRecord *));
	recordCnt = 0;
	}

// *****************************************************************************
// * PBS_JobArray::sort  :
// *	This method re-sorts the PBS_JobArray using the specified sort function.
// *****************************************************************************
void PBS_JobArray::sort ( PBS_JobRecordSorter Sorter )
	{
	qsort(records, recordCnt, sizeof(PBS_JobRecord *), (int (*)(const void *, const void *))Sorter);
	}	

// *****************************************************************************
// * PBS_JobArray::addJob  :
// *	This method adds a new job to the array. The PBS_JobRecord is initially
// *	installed in the PBS_JobCollection, then a pointer to the PBS_JobRecord is
// *	placed into the PBS_JobArray.
// *****************************************************************************
PBS_JobRecord * PBS_JobArray::addJob ( PBS_JobRecord * job )
	{
	PBS_JobRecord * oldJob = NULL;
	
	if(job && job->getJobID())
		{
		int found = 0;
		int idx   = 0;
		oldJob    = PBS_JobCollection::addJob(job);

		if(oldJob)
			{
			for(idx=0; idx<recordCnt; )
				{
				if(records[idx] == oldJob)
					{
					memmove(&records[idx], &records[idx+1], (recordCnt-idx-1)*sizeof(PBS_JobRecord *));
					records[--recordCnt] = NULL;
					}
				else idx++;
				}
			}
		
		if(recordCnt+1 > arraySize)
			{
			records = (PBS_JobRecord **)realloc(records, 2*arraySize*sizeof(PBS_JobRecord *));
			memset(&records[arraySize], 0, arraySize*sizeof(PBS_JobRecord *));
			arraySize*=2;
			}
		records[recordCnt++] = job; 
		}

	return oldJob;
	}
	
// *****************************************************************************
// * PBS_JobArray::delJob  :
// *	This methods removes a PBS_JobRecord from the PBS_JobCollection and then 
// *	removes it from the PBS_JobArray.  The PBS_JobRecord is not deleted, and 
// *	becomes the property of the caller.
// *****************************************************************************
PBS_JobRecord * PBS_JobArray::delJob ( char * job )
	{
	PBS_JobRecord * oldJob = PBS_JobCollection::delJob(job);
	if(oldJob)
		{
		int idx;
		for(idx=0; idx<recordCnt && records[idx]!=oldJob; idx++);
		if(idx<recordCnt)
			{
			memmove(&records[idx], &records[idx+1], (recordCnt-idx-1)*sizeof(PBS_JobRecord *));
			records[--recordCnt]=NULL;
			}
		}
	return oldJob;
	}
	

// *****************************************************************************
// * PBS_JobArray::asciiDump  :
// *	This method writes the ascii contents of the object to the specified
// *	file pointer.
// *****************************************************************************
void PBS_JobArray::asciiDump ( FILE * fp, char * prefix )
	{
	fprintf(fp, "\n%sReporting Jobs for %s\n\n", prefix, name);
	for(int idx=0; idx<recordCnt; idx++)
		{
		if(records[idx]) records[idx]->asciiDump(fp, prefix, !idx);
		else fprintf(fp, "%sWARNING: Improper NULL Record in PBS_JobArray %s\n", prefix, getName());
		}
	fprintf(fp, "\n");
	}

// *****************************************************************************
// * PBS_JobArray::operator [] :
// *	This a safe index operator for the class.
// *****************************************************************************
const PBS_JobRecord * PBS_JobArray::operator [] ( size_t idx )
	{
	return (idx<recordCnt)?records[idx]:NULL; 
	}
	
	
