// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the methods that are defined in the 
// 	PBS_JobCollection.h header file.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include "PBS_JobCollection.h"
#include "PBS_JobRecord.h"

// *****************************************************************************
// * PBS_JobCollection::PBS_JobCollection:
// *	Constructor for the PBS_JobCollection object.  Initializes the internal
// *	variables (name) and calls the constructor for embedded objects
// *	(jobHash).
// *****************************************************************************
PBS_JobCollection::PBS_JobCollection ( const char * Name, size_t hashSize )
	: name((Name && *Name)?strdup(Name):NULL), jobHash(hashSize)
	{
	}
	

// *****************************************************************************
// * JobColleciton::~PBS_JobCollection :
// *	Destructor for the class.  It uses free to release the memory allocated
// *	to 'name' using strdup, and then walks through each job in the jobHash
// *	and deletes it.  The jobHash embedded object will be auto-deleted.
// *****************************************************************************
PBS_JobCollection::~PBS_JobCollection ( void )
	{
	if(name) free(name);
	clear();
	}
	
// *****************************************************************************
// * PBS_JobCollection::clear :
// *	This method clears the class by deleteing all of the PBS_JobRecord objects
// *	from the JobHash structure.  The name string is left intact.
// *****************************************************************************
void PBS_JobCollection::clear ( void )
	{
	PBSStringHashIter iter(jobHash);
	PBS_JobRecord       * job;

	while((job = (PBS_JobRecord *)iter.data()))
		{
		iter++;
		delete job;
		}
	}

// *****************************************************************************
// * PBS_JobCollection::addJob :
// *	This method will add a new PBS_JobRecord object to the jobHash. If a record
// *	with a matching jobID already exists (and it is not the same record),
// *	the existing record will be returned to the caller for deletion.  NULL
// *	will be returned if an old record does not exist.
// *
// *	It is atypical that an old record should exist.  If an old record is
// *	found, it indicates a programming error somewhere else in the code.
// *****************************************************************************
PBS_JobRecord * PBS_JobCollection::addJob ( PBS_JobRecord * job )
	{
	PBS_JobRecord * oldJob = NULL;
	
	// *********************************************************************
	// * Before proceeding, insure that the caller provided a valid job.
	// *********************************************************************
	if(job && job->getJobID())
		{
		// *************************************************************
		// * Look for an existing job that has the same job ID.
		// *************************************************************
		oldJob = (PBS_JobRecord *)jobHash.find((char *)job->getJobID());
		
		// *************************************************************
		// * If the old Job is the same as the new job, do nothing.
		// *************************************************************
		if(oldJob==job) 
			{
			oldJob=NULL;
			}
		// *************************************************************
		// * If there was no old job or the old job and new job differ.
		// *************************************************************
		else	{
			// *****************************************************
			// * If there was an old job, remove it from the jobHash
			// *****************************************************
			if(oldJob) jobHash.remove((char *)oldJob->getJobID());
			
			// *****************************************************
			// * Insert the new job using the jobID as the key.
			// *****************************************************
			jobHash.insert((char *)job->getJobID(), job);
			}
		}
	
	// *********************************************************************
	// * Return the old job pointer to be deleted if it is Non-NULL.
	// *********************************************************************
	return oldJob;
	}
	
// *****************************************************************************
// * PBS_JobCollection::asciiDump :
// *	This method performs a text dump of the object contents to a file.  This
// *	method is provided for debugging and other general purpose use.
// *****************************************************************************
void PBS_JobCollection::asciiDump ( FILE * fp, char * prefix )
	{
	PBSStringHashIter iter(jobHash);
	PBS_JobRecord * job;
	int header = 0;
	
	while((job = (PBS_JobRecord *)iter.data()))
		{
		job->asciiDump(fp, prefix, !header++);
		iter++;
		}
	}
