// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This is a collection of classes that allows the caller to use and 
// 	manipulate a collection of profile strings that are stored in disk
// 	files.
// 	
//	The structure of these data files was originally introduced as 
//	part of the Microsoft Windows operating system.  Over the past
//	15 years, I have adapted the structure, syntax and functionality
//	somewhat to make this system more applicable to product development.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <PBS_Hash.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

// *****************************************************************************
// * class ProfileManager :
// *	This class contains a list of profile files and a pointer to their
// *	associated Profile object.
// *****************************************************************************
class ProfileManager
{
private:
	PBSStringHash tables;
	
public:
	inline               ProfileManager    (void);
	inline              ~ProfileManager    (void);
	inline void          reload            (void);
	inline void          reload            (const char * path);
	inline const char *  getProfileString  (const char * path, const char * header, const char * name, const char * defValue=NULL);
	inline const char ** getProfileHeaders (const char * path);
	inline const char ** getProfileEntries (const char * path, const char * header);
	inline void          setProfileString  (const char * path, const char * header, const char * name, const char * value, int local = 0 );
	inline void          asciiDump         (FILE * fp=stdout, char * prefix=NULL);
};


// *****************************************************************************
// * class Profile :
// *	This class represents a single profile file that contains entries of
// *	the form.
// *	
// *	# comment
// *	[header]
// *	entry1=value1
// *	entry2=value2
// *
// *	The entire file is loaded at initialization, however, the user may 
// *	reload the file at any time by calling the reload method. Formatted
// *	entries are requested from the table by using the command 
// *	getProfileString and specifying the header and entry desired.
// *	The user may write a profile entry by using the command 
// *	setProfileString with the header and entry desired.  If the user lacks
// *	sufficient access to write to the file, the data will be updated
// *	locally, however, it will not be written to the disk.
// *****************************************************************************
class Profile 
{
private:
	// *********************************************************************
	// * char * path:
	// *	This is the path to the file containing the profile information.
	// *********************************************************************
	char * path;
	
	// *********************************************************************
	// * class Entry :
	// *	This object contains an entry name, entry string and a flag
	// *	inidcating whether it is a local or global value.  Values with
	// *	the local flag set will not be modified when the file is 
	// *	reloaded and are not written to disk when they are modified 
	// *	locally. Global values will be reloaded from disk each time
	// *	the caller executes a reload command.
	// *********************************************************************
	class Entry
	{
	private:
		char * name;
		char * value;
		int    local;
	
	public:
		inline Entry                ( const char * Name, const char * Value, int Local=0 );
		inline ~Entry               ( void );
		inline void update          ( const char * Value, int Local=0 );
		inline void updateFromDisk  ( const char * Value );
		inline const char *getName  ( void );
		inline const char *getValue ( void );
		inline int         getLocal ( void );
	};
		
	PBSStringHash headers;

	inline int          writeEntry       ( const char * header, const char * name, const char * value );
	inline char *       removeExtraSpaces(const char * in, char * out);	

public:
	inline               Profile           ( const char * Path );
	inline              ~Profile           ( void );
	inline const char  * getPath           ( void );
	inline const char  * getProfileString  ( const char * header, const char * name );
	inline const char ** getProfileHeaders ( void );
	inline const char ** getProfileEntries ( const char * header );
	inline void          setProfileString  ( const char * header, const char * name, const char * value, int local );
	inline void          asciiDump         ( FILE * fp, char * prefix );
	inline void          reload            ( void );
	
};


// *****************************************************************************
// * ProfileManager::ProfileManager :
// *	Constructor for the ProfielList class it initializes all of the
// *	local variables.
// *****************************************************************************
inline ProfileManager::ProfileManager ( void )
	: tables()
	{
	}
	
// *****************************************************************************
// * ProfileManager::~ProfileManager :
// *	Destructor for the ProfileManager class it deletes any memory that was
// *	allocated during the objects use.
// *****************************************************************************
inline ProfileManager::~ProfileManager ( void )
	{
	PBSStringHashIter iter (tables);
	Profile * table;
	
	for(iter.first(); (table = (Profile *)iter.data())!=NULL; iter++)
		{
		delete table;
		}
	}
	
// *****************************************************************************
// * ProfileManager::getProfileString :
// *	Using a path, header and entry name, this method will retrieve a 
// *	specific string from the database of entries. Note: Even though the
// *	data in the file may have changed externally, these changes will not
// *	be reflected in the local database until the user calls the reload
// *	method.
// *****************************************************************************
inline const char * ProfileManager::getProfileString (
	const char * path, 
	const char * header, 
	const char * name, 
	const char * defValue)
	{
	const char * result = NULL;
	Profile * table = NULL;
	
	if(path && *path)
		{
		if((table = (Profile *)tables.find(path))==NULL)
			{
			table = new Profile(path);
			tables.insert(table->getPath(), table);
			}
		result = table->getProfileString(header, name);
		}
	return (result && *result)?result:defValue;
	}

// *****************************************************************************
// * ProfileManager::getProfileHeaders :
// *	This method will retrieve an array of all of the headers that have
// *	been registered in the specified profile.  The array becomes the 
// *	property of the caller and must be released using free when it is no 
// *	longer needed. The strings themselves remain the property of the
// *	underlying object and should not be modified or deleted.
// *****************************************************************************
inline const char ** ProfileManager::getProfileHeaders ( const char * path )
	{
	const char ** result = NULL;
	Profile     * table = NULL;

	if( path && *path && (table = (Profile *)tables.find(path))!=NULL)
		{
		result = table->getProfileHeaders ();
		}

	return result;	
	}
	
// *****************************************************************************
// * ProfileManager::getProfileEntries :
// *	This method will retrieve an array of all of the entries associated
// *	with a specific header that have been registered in a Profile.  The 
// *	array becomes the property of the caller and must be released using free
// *	when it is no longer needed. The strings themselves remain the property 
// *	of the underlying object and should not be modified or deleted.
// *****************************************************************************
inline const char ** ProfileManager::getProfileEntries ( 
	const char * path,
	const char * header )
	{
	const char ** result = NULL;
	Profile     * table = NULL;

	if( path && *path && (table = (Profile *)tables.find(path))!=NULL)
		{
		result = table->getProfileEntries (header);
		}

	return result;	
	}
	
	
// *****************************************************************************
// * ProfileManager::setProfileString :
// *	Using a path, header and entry name, this method will update the 
// *	entry in the database. The local flag is passed to the method that 
// *	actually performs the work. If the local flag is 0 (FALSE), the that
// *	method will also attempt to update the entry in the data file.
// *
// *	See Profile::setProfileString for details.
// *****************************************************************************
inline void ProfileManager::setProfileString (
	const char * path, 
	const char * header, 
	const char * name,
	const char * value,
	int          local )
	{
	Profile * table = NULL;

	if(path && *path)
		{
		if((table = (Profile *)tables.find(path))==NULL)
			{
			table = new Profile(path);
			tables.insert(table->getPath(), table);
			}
		table->setProfileString(header, name, value, local);
		}
	}
	
// *****************************************************************************
// * ProfileManager::reload :
// *	This method will reload the contents of the online database from the
// *	disk contents.  Any variables that were declared as local when they
// *	were inserted earlier will remain unchanged.
// *****************************************************************************
inline void ProfileManager::reload ( void )
	{
	PBSStringHashIter iter (tables);
	Profile * table;
	iter.first();
	while((table = (Profile *)iter.data())!=NULL)
		{
		table->reload();
		iter++;
		}
	}

// *****************************************************************************
// * ProfileManager::reload :
// *	This method will reload the file specified in the path variable.
// *****************************************************************************
inline void ProfileManager::reload ( const char * path )
	{
	Profile * table = NULL;
	if(path && *path)
		{
		if((table = (Profile *)tables.find(path))==NULL)
			{
			table = new Profile(path);
			tables.insert(table->getPath(), table);
			}
		table->reload();
		}
	}
	
// *****************************************************************************
// * ProfileManager::asciiDump :
// *	This method generates an ASCII report of the contents of all databases
// *	currently registered with the ProfileManager.  This method is provided
// *	purely for debugging and summary information.
// *****************************************************************************
void ProfileManager::asciiDump ( FILE * fp, char * prefix )
	{
	PBSStringHashIter iter(tables);
	Profile * table;

	for(iter.first(); (table=(Profile *)iter.data())!=NULL; iter++)
		{
		table->asciiDump(fp, prefix);
		}
	}

// *****************************************************************************
// * Profile::Entry::Entry :
// *	This is the constructor for the Entry subclass of the Profile
// *	class.  It will make copies of the variables that were passed to it 
// *	as arguments.
// *****************************************************************************
inline Profile::Entry::Entry  (const char *Name, const char *Value, int Local)
	: name(strdup(Name)),
	  value((Value && *Value)?strdup(Value):NULL),
	  local(Local)
	{
	}

// *****************************************************************************
// * Profile::Entry::~Entry :
// *	This is the destructor for the Entry subclass of the Profile 
// *	class.  It will free all of the strings that were allocated during its 
// *	use.
// *****************************************************************************
inline Profile::Entry::~Entry ( void )
	{
	if(name)  free(name);
	if(value) free(value);
	}

// *****************************************************************************
// * Profile::Entry::update :
// *	This is the update method of the Entry subclass of the Profile
// *	class.  It will update the data stored in an existing Entry object
// *	to reflect the newer, user specified value.  
// *****************************************************************************
inline void Profile::Entry::update ( const char * Value, int Local )
	{
	if(value) free(value);
	value = ((Value && *Value)?strdup(Value):NULL);
	local = Local;
	}

// *****************************************************************************
// * Profile::Entry::updateFromDisk :
// *	This method is called when the data is being reloaded from disk.  If the
// *	local flag is set (non-zero) that indicates that the variable in this
// *	entry is a local value that will not be reflected in the disk file.
// *	Consequently, all local variables are exempt from being updated from
// *	the disk file during the reload process.
// *****************************************************************************
inline void Profile::Entry::updateFromDisk  ( const char * Value )
	{
	if(!local) update(Value);
	}

// *****************************************************************************
// * Profile::Entry::getValue :
// *	This is a convenience function that allows the Profile to 
// * 	retrieve the values stored in the Entry object.
// *****************************************************************************
inline const char * Profile::Entry::getValue ( void )
	{
	return (const char *)value;
	}

// *****************************************************************************
// * Profile::Entry::getName :
// *	This is a convenience function that allows the Profile to retrieve
// *	the name of an Entry sub-object.
// *****************************************************************************
inline const char * Profile::Entry::getName ( void )
	{
	return (const char *)name;
	}
	
// *****************************************************************************
// * Profile::Entry::getLocal :
// *	This is a convenience function that allows the Profile object to 
// *	retrieve the local flag that is stored in the Entry sub-object.
// *****************************************************************************
inline int Profile::Entry::getLocal ( void )
	{
	return local;
	}
	

// *****************************************************************************
// * Profile::Profile :
// *	This is the constructor for the Profile object. It makes a copy
// *	of the variables that were passed to it as arguments and then 
// *	initializes all internal variables.  The method will call the reload 
// *	method to populate itself with data that is located in the disk file
// *	specified by Path.
// *****************************************************************************
inline Profile::Profile ( const char * Path )
	: path      ((Path && *Path)?strdup(Path):NULL),
	  headers   (256, 1)
	{
	reload();
	}
	
// *****************************************************************************
// * Profile::~Profile :
// *	This is the dstructor for the Profile class. It will free the 
// *	path string tht was allocated earlier using the strdup command.
// *****************************************************************************
inline Profile::~Profile ( void )
	{
	if(path) free(path);

	PBSStringHashIter iter (headers);
	PBSStringHash * entries;

	for(iter.first(); (entries = (PBSStringHash *)iter.data())!=NULL; iter++)
		{
		PBSStringHashIter eIter(*entries);
		Entry * entry;
		
		for(eIter.first(); (entry = (Entry *)eIter.data())!=NULL; eIter++)
			{
			delete entry;
			}
		delete entries;
		}
	}
	
// *****************************************************************************
// * Profile::getPath :
// *	This is a convenience function that allows the caller to retrieve a
// *	pointer to the path variable. The path variable indicates the location
// *	of the database file on disk.
// *****************************************************************************
inline const char * Profile::getPath ( void )
	{
	return (const char *)path;
	}
	
// *****************************************************************************
// * Profile::getProfileString:
// *	This method allows the caller top retrieve a string value from the 
// *	online database using the header and entry name.  Note that even though
// *	the physical database may have changed on disk, the caller will not see
// *	the updated data until he uses the reload command.  Even after the 
// *	reload method is caller, any data entries that were marked as local 
// *	when they were created will not be updated from the data on disk.
// *
// *	See the Profile::reload method for more details.
// *****************************************************************************
inline const char * Profile::getProfileString ( 
	const char * header, 
	const char * name )
	{
	Entry         * entry   = NULL;
	PBSStringHash * entries = NULL;
	char            fmtHeader[256];
	char            fmtName  [256];

	if(header && *header && strlen(header)<256 &&
	   name   && *name   && strlen(name)<256   &&
	   *removeExtraSpaces(header, fmtHeader)   &&
	   *removeExtraSpaces(name, fmtName))
		{
		if((entries = (PBSStringHash *)headers.find(fmtHeader))!=NULL)
			{
			entry = (Entry *)entries->find(fmtName);
			}
		}
	return (entry?entry->getValue():NULL);
	}
	
// *****************************************************************************
// * Profile::getProfileHeaders::
// *	This method will return an array of strings containing all of the 
// *	headers that have been stored in this Profile object.  The array becomes
// *	the property of the caller, however, the strings remain the property of
// *	the Profile object. Note, since the string array is being allocated 
// *	using the malloc and realloc methods, the caller should use free to 
// *	release the memory.
// *****************************************************************************
inline const char ** Profile::getProfileHeaders ( void )
	{
	size_t         resultMax = 0;
	size_t         resultCnt = 0;
	const char **  result    = NULL;
	const char   * header    = NULL;
	
	PBSStringHashIter iter(headers);
	for(iter.first(); (header=iter.key())!=NULL; iter++)
		{
		if(resultCnt==0)
			{
			resultMax = 16;
			result = (const char **)malloc(resultMax*sizeof(char *));
			}
		if((resultCnt+2) >= resultMax)
			{
			resultMax*=2;
			result = (const char **)realloc(result, resultMax*sizeof(char *));
			}
		result[resultCnt++] = header;
		}
	if(result) result[resultCnt]=NULL;
	return result;
	}

// *****************************************************************************
// * Profile::getProfileEntries::
// *	This method will return an array of strings containing all of the 
// *	entries that have been stored in this Profile object under the specified
// *	header.  The array becomes the property of the caller, however, the 
// *	strings remain the property of the Profile object. Note, since the 
// *	string array is being allocated using the malloc and realloc methods, 
// *	the caller should use free to release the memory.
// *****************************************************************************
inline const char ** Profile::getProfileEntries ( const char * header )
	{
	size_t          resultMax = 0;
	size_t          resultCnt = 0;
	const char   ** result    = NULL;
	PBSStringHash * entries   = NULL;
	char            fmtHeader[256];
	char            fmtName  [256];

	if(header && *header && strlen(header)<256 &&
	   *removeExtraSpaces(header, fmtHeader)   &&
	   (entries = (PBSStringHash *)headers.find(fmtHeader))!=NULL)
		{
		PBSStringHashIter iter(*entries);
		for(iter.first(); (header=iter.key())!=NULL; iter++)
			{
			if(resultCnt==0)
				{
				resultMax = 16;
				result = (const char **)malloc(resultMax*sizeof(char *));
				}
			if((resultCnt+2) >= resultMax)
				{
				resultMax*=2;
				result = (const char **)realloc(result, resultMax*sizeof(char *));
				}
			result[resultCnt++] = header;
			}
		}
	if(result) result[resultCnt]=NULL;
	return result;
	}


// *****************************************************************************
// * Profile::setProfileString :
// *	This method will write the user specified string to the database using
// *	the specified header and entry name.  If the local flag is TRUE 
// *	(non-zero), then only the online database will be updated.  If the 
// *	the local flag is FALSE (0), then the method will attempt to write
// *	the data to the database on disk.  In the event that the user does not
// *	have sufficient access to update the database, then the entry will be
// *	automatically marked as local.
// *
// *	See the Profile::writeEntry method for more details.
// *****************************************************************************
inline void Profile::setProfileString ( 
	const char * header,
	const char * name, 
	const char * value, 
	int local )
	{
	char fmtHeader[256];
	char fmtName  [256];

	if(header && *header && strlen(header)<256 &&
	   name   && *name   && strlen(name)<256   &&
	   *removeExtraSpaces(header, fmtHeader)   &&
	   *removeExtraSpaces(name, fmtName))
		{
		char * fmtValue = new char[value?strlen(value)+1:1];
		removeExtraSpaces(value, fmtValue);
		
		if(!local) local=writeEntry(fmtHeader, fmtName, fmtValue);
	
		Entry         * entry   = NULL;
		PBSStringHash * entries = NULL;

		if((entries=(PBSStringHash *)headers.find(fmtHeader))==NULL)
			{
			headers.insert(fmtHeader, (entries=new PBSStringHash()));
			}
		if((entry = (Entry *)entries->find(fmtName))==NULL)
			{
			entry = new Entry(fmtName, value, local);
			entries->insert(entry->getName(), entry);
			}
		else entry->update(value, local);
		
		delete[] fmtValue;
		}
	}

// *****************************************************************************
// * ProfileManager::asciiDump :
// *	This method generates an ASCII report of the contents of all databases
// *	currently registered with the ProfileManager.  This method is provided
// *	purely for debugging and summary information.
// *****************************************************************************
void Profile::asciiDump ( FILE * fp, char * prefix )
	{
	char   blank = 0;
	char * prefixString = prefix?prefix:&blank;

	fprintf(fp, "%s---------------------------------------------------------------------\n", prefixString);
	fprintf(fp, "%sProfile Report for: %s\n", prefixString, path);
	fprintf(fp, "%s---------------------------------------------------------------------\n", prefixString);
	
	PBSStringHashIter hIter(headers);
	PBSStringHash * entries;
	
	for(hIter.first(); (entries=(PBSStringHash *)hIter.data())!=NULL; hIter++)
		{
		fprintf(fp, "%s\n[%s]\n", prefixString, hIter.key());
		
		PBSStringHashIter eIter(*entries);
		Entry * entry;
		
		for(eIter.first(); (entry=(Entry *)eIter.data())!=NULL; eIter++)
			{
			fprintf(fp, "%s%s=%s\n", prefixString, entry->getName(), entry->getValue()?entry->getValue():"\0");
			}
		}
	}

// *****************************************************************************
// * Profile::reload :
// *	This method will cause the Profile to reload data from the 
// *	database on disk.  All data that is marked as non-local will be updated
// *	from the disk file.  Any data items that were specified as local (either
// *	intentionally or because the user lacked sufficient access to write the
// *	changes to disk) WILL NOT be updated --- and will remain unchanged
// *	even though the physical database may differ.
// *****************************************************************************
inline void Profile::reload( void )
	{
	// *********************************************************************
	// * Open the input file for reading. If this is successful, process the
	// * contents of the file.
	// *********************************************************************
	int fd = open(path, O_RDONLY);
	
	if(fd>0)
		{
		// *************************************************************
		// * Determine the length of the file and allocate a sufficient
		// * buffer.
		// *************************************************************
		size_t size=lseek(fd, 0L, SEEK_END);
		char * buf = new char[size+2];
		*buf='\n';
		
		// *************************************************************
		// * Read the contents of the file into the buffer and close it.
		// *************************************************************
		lseek(fd, 0L, SEEK_SET);
		buf[read(fd, buf+1, size)+1] = 0;
		close(fd);
		
		PBSStringHash  * entries;
		Entry          * entry;
		
		char *ptr = buf;
		char *nextPtr, *linePtr, *valuePtr, *cmtPtr, *utilPtr;
		
		// *************************************************************
		// * Walk through the buffer stopping at each line that begins
		// * with a '[' charcter.
		// *************************************************************
		for(ptr = strstr(buf, "\n["); ptr!=NULL; ptr=nextPtr)
			{
			// *****************************************************
			// * Set ptr to the first character in the section head.
			// *****************************************************
			ptr+=2;

			// *****************************************************
			// * Set nextPtr to the beginning of the next section...
			// * If the section is found, NULL terminate the buffer
			// * at the nextPtr to provide a defined region.
			// *****************************************************
			if((nextPtr = strstr(ptr, "\n["))!=NULL) *nextPtr = 0;

			// *****************************************************
			// * Find the terminating character for the heading.
			// *****************************************************
			if((utilPtr = strchr(ptr, ']'))!=NULL)
				{
				// *********************************************
				// * NULL terminate the heading string and use
				// * this value (currently identified as ptr
				// * to locate the heading in the headers 
				// * object. If it doesn't exist create it.
				// *********************************************
				*utilPtr = 0;
				if((entries=(PBSStringHash *)headers.find(ptr))==NULL)
					{
					entries = new PBSStringHash();
					headers.insert(ptr, entries);
					}
				ptr = utilPtr+1;

				// *****************************************************
				// * Move to the beginning of each line and process
				// * each entry in the section.
				// *****************************************************
				for(ptr=strchr(ptr, '\n'); ptr && *ptr=='\n'; ptr=linePtr)
					{
					// *********************************************
					// * Step forward to the first character of the 
					// * line.
					// *********************************************
					ptr++;
					
					// *********************************************
					// * Set linePtr to the end of the current line
					// * and NULL terminate it to provide a defined
					// * region.
					// *********************************************
					if((linePtr = strchr(ptr, '\n'))!=NULL) *linePtr = 0;

					// *********************************************
					// * If a comment character exists on the line,
					// * NULL terminate it to reduce the defined
					// * region.
					// *********************************************
					if((cmtPtr = strchr(ptr, '#'))!=NULL) *cmtPtr = 0;
				
					// *********************************************
					// * Walk through all of the leading whitespace.
					// *********************************************
					while(isspace(*ptr) && *ptr) ptr++;
					
					// *********************************************
					// * If the line has not been commented out, 
					// * locate the equals sign on the line. If the
					// * '=' is at the beginning of the line, then
					// * the entry is invalid and should be ignored.
					// *********************************************
					if(*ptr && (utilPtr=strchr(ptr, '='))!=NULL && utilPtr>ptr)
						{
						// *************************************
						// * Set the valuePtr to the first char
						// * after the '=' and walk through the
						// * empty spaces.
						// *************************************
						for(	valuePtr = utilPtr+1; 
							*valuePtr && isspace(*valuePtr); 
							valuePtr++);

						// *************************************
						// * NULL terminate the entry name at 
						// * the equal sign and remove trailing
						// * blank space.
						// *************************************
						do	{
							*(utilPtr--) = 0;
							} while(utilPtr>=ptr && isspace(*utilPtr));
							
						// *************************************
						// * Set the utilPtr to the end of the
						// * line and remove trailing blank 
						// * space after the value.
						// *************************************
						utilPtr = valuePtr+strlen(valuePtr);
						do	{
							*(utilPtr--)=0;
							} while(isspace(*utilPtr) && utilPtr>=valuePtr);

						// *************************************
						// * If an Entry object doesn't exist
						// * for this entry, create one and 
						// * install the new data.  Otherwise,
						// * update the existing entry object.
						// *************************************
						if((entry = (Entry *)entries->find(ptr))==NULL)
							{
							entry = new Entry(ptr, valuePtr);
							entries->insert(entry->getName(), entry);
							}
						else entry->updateFromDisk(valuePtr);
						}
					// *********************************************
					// * If the linePtr is valid, restore the '\n'
					// * character.  The ptr variable will be 
					// * updated to this position at the end of the
					// * loop.
					// *********************************************
					if(linePtr) *linePtr = '\n';
					}
				}
			
			// *****************************************************
			// * If the nextPtr is valid, restore the '\n' character
			// * The ptr variable will be updated to this position
			// * at the end of the loop.
			// *****************************************************
			if(nextPtr) *nextPtr = '\n';
			}

		// *************************************************************
		// * Delete the buffer that was allocated to hold the data
		// * from the file.
		// *************************************************************
		delete[] buf;
		}
	}
	
// *****************************************************************************
// * Profile::writeEntry :
// *	This method is called to write an individual entry to the physical
// *	database on disk.  The individual record approach is used to protect
// *	comments that the system administrator may have introduced into the
// *	database file.
// *
// *	In the event that the data was successfully written to disk, a value of
// *	0 will be returned.  If the user lacked sufficient access to write the 
// *	data to disk, then a value of -1 is returned... This value is then
// *	applied to the local variable of the data entry.
// *
// *	If the local variable within a data entry is set to TRUE, then that 
// *	entries value takes precedence over any changes in the physical 
// *	database. Consequently, when the database is reloaded using the 
// *	Profile::reload command, any entry that is marked as local will
// *	not be updated with the new value.
// *****************************************************************************
inline int Profile::writeEntry ( 
	const char * header, 
	const char * name, 
	const char * value )
	{
	int result = 0;
	FILE * fp;
	
	if(access(path, F_OK)) fp = fopen(path, "w");
	else fp = fopen(path, "r+");
	
	if(fp!=NULL && header && *header && name && *name)
		{
		// *************************************************************
		// * Determine the size of the file, allocate a buffer of 
		// * sufficient size and read the entire file into the buffer.
		// * Note: I am prepending a carriage return to the beginning
		// * of the buffer so that any heading will always be prefixed
		// * by a carriage return. The last character of the document
		// * should also be a carriage return.
		// *************************************************************
		fseek(fp, 0L, SEEK_END);
		size_t size = ftell(fp), newSize = 0;
		char * buf = new char[size+3];
		*buf='\n';
		fseek(fp, 0L, SEEK_SET);
		fread(buf+1, 1, size, fp);
		if(buf[size]!='\n')
			{
			buf[size+1]='\n';
			buf[size+2]=0;
			}
		else buf[size+1] = 0;

		// *************************************************************
		// * Generate the header search string and search for the header
		// * in the file. If the header cannot be found, then seek
		// * to the end of the file and append the new header and entry.
		// *************************************************************
		char * ptr, *nextHeader, *lastLine=NULL, *currLine, *nextLine, *utilPtr;
		char   cNextLine;
		char * headerBuf = new char[strlen(header)+4];
		sprintf(headerBuf, "\n[%s]", header);
		if((ptr = strstr(buf, headerBuf))==NULL)
			{
			fseek(fp, 0L, SEEK_END);
			fprintf(fp, "\n[%s]\n%s=%s\n\n", 
				header, name, (value && *value)?value:"\0");
			newSize = ftell(fp);
			}
		// *************************************************************
		// * In the event that the header was located, then locate the
		// * last valid entry of this section.  NULL terminate the 
		// * buffer at this point to provide an isolated search region.
		// *************************************************************
		else	{
			for(++ptr; *ptr!='\n' && *ptr; ptr++);
			if((nextHeader = strstr(ptr, "\n["))) lastLine=nextHeader;
			else lastLine = buf+strlen(buf)-1;
			while(lastLine>ptr && *lastLine=='\n') lastLine--;
			
			int lastFound = 0;
			while(!lastFound)
				{
				while(lastLine>ptr && *lastLine!='\n') lastLine--;
				
				if(lastLine<ptr)
					{
					lastLine  = ptr;
					lastFound = 1;
					}
				else	{
					utilPtr = lastLine-1;
					while(*lastLine && isspace(*(++lastLine)) && *lastLine!='\n');
					// *****************************
					// * Comments are assumed to be
					// * associated with the next 
					// * header - not the previous
					// * line.
					// *****************************
					if(*lastLine=='\n' || *lastLine=='#')
						{
						lastLine = utilPtr;
						}
					else	{
						while(*lastLine && *lastLine!='\n') lastLine++;
						lastFound = 1;
						}
					}
				}
			*lastLine = 0;
							
			while(*ptr=='\n' && newSize==0)
				{
				// *********************************************
				// * Find the beginning of the nextLine or a
				// * trailing comment character and NULL
				// * terminate the buffer at this point.  That 
				// * will provide an isolated search region for
				// * the entry.
				// *********************************************
				for(nextLine=++ptr; 
				   *nextLine && *nextLine!='#' && *nextLine!='\n'; 
				    nextLine++);
				cNextLine = *nextLine;
				*nextLine = 0;
				
				char *namePtr, *valuePtr;
				size_t nameLen=0, valueLen;
				
				for(namePtr=ptr; 
				   *namePtr && isspace(*namePtr); 
				    namePtr++);
				
				// *********************************************
				// * Locate the '=' character on the line and 
				// * use that to identify the length of the name
				// *********************************************
				if((utilPtr=strchr(namePtr, '='))!=NULL)
					{
					valuePtr = utilPtr+1;
					while(utilPtr>namePtr && isspace(*(--utilPtr)));
					nameLen = (size_t)(utilPtr-namePtr)+1;
					
					while(*valuePtr && isspace(*valuePtr)) valuePtr++;
					for(utilPtr=nextLine-1; 
					    utilPtr>valuePtr && isspace(*utilPtr); 
					    utilPtr--);
					valueLen = (size_t)(utilPtr-valuePtr)+1;
					}
				// *********************************************
				// * Restore the character that was NULLed to
				// * temporarily terminate the line.
				// *********************************************
				*nextLine = cNextLine;
				
				// *********************************************
				// * If the namePtr contains data and its value
				// * is the same as the name parameter, then we
				// * have found the location for the update and
				// * we should write the new record here and
				// * then write any data that follows it.
				// *********************************************
				if(nameLen && strncmp(name, namePtr, nameLen)==0)
					{
					fseek(fp, (ptr-buf)-1, SEEK_SET);
					fprintf(fp, "%s=%s", name, value);
					*lastLine = '\n';
					if(*nextLine)
						{
						if(*nextLine=='#') fprintf(fp, " ");
						fwrite(nextLine, 1, strlen(nextLine), fp);
						}
					else fprintf(fp, "\n");
					newSize = ftell(fp);
					}
				// *********************************************
				// * If the name is not a match, then advance to
				// * the next line and continue.
				// *********************************************
				else for(ptr=nextLine; *ptr && *ptr!='\n'; ptr++);
				}
			// *****************************************************
			// * If the entry was not located within the section, 
			// * then reposition to the last line of the section and
			// * create the new entry there.
			// *****************************************************
			if(newSize==0)
				{
				*lastLine = '\n';
				fseek(fp, (lastLine-buf)-1, SEEK_SET);
				fprintf(fp, "\n%s=%s", name, value);
				fwrite(lastLine, 1, strlen(lastLine), fp);
				newSize = ftell(fp);
				}
			}
		
		fclose(fp);
		truncate(path, newSize);
		delete[] buf;
		delete[] headerBuf;
		}
	else result = -1;
	
	return result;
	}
	
// *****************************************************************************
// * removeExtraSpaces :
// *	This method will remove leading and trailing spaces from the input
// * 	string and place the resulting string in the output string. It is 
// * 	assumed that the user has allocated a string of sufficient size to hold
// *	the resulting string.
// *****************************************************************************
inline char * Profile::removeExtraSpaces ( const char * input, char * output )
	{
	char   blank = 0;
	char * ptr   = output;

	if(!input) input = &blank;
	while(isspace(*input)) input++;
	while(*input) 
		{
		if(*input=='#' || *input=='\n' || *input=='[' || *input==']')
			*ptr = ' ';
		else	*ptr = *input;

		input++;
		ptr++;
		}
	*ptr = 0;

	ptr = output+strlen(output);
	while(ptr>=output && (*ptr==0 || isspace(*ptr))) *(ptr--)=0;

	return output;	
	}
	
