// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Queue.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Queue
// *              class.  This class is generally contained by a PBS_Server 
// *              class which provides populates it and provides access to 
// *              its components.
// *****************************************************************************

#ifndef _PBS_QUEUE_H_
#define _PBS_QUEUE_H_ 1

#include <PBS_Job.h>

// *****************************************************************************
// * class PBS_Queue :
// *    This class provides the mechanisms required to read all available
// *    status/configuration information from a PBS queue.  It is assumed
// *    that this information will be used later for assessing conditions
// *    when scheduling jobs.
// *
// *    The following data items are defined in this class:
// *
// *         name            : The name of the queue.
// *         queue_type      : An enumerated value identifying the type of the 
// *                           queue.  It can be either EXECUTION or ROUTING.
// *         enabled         : A boolean flag indicating whether the queue 
// *                           is enabled.
// *         started         : A boolean flag indicating whether the queue has 
// *                           been started.
// *         priority        : An integer containing the priority of the queue.
// *         total_jobs      : An integer containing the total number of jobs 
// *                           currently in the queue.
// *         state_count     : An array of integers identifing the number of
// *                           jobs in each of six states. The integers are 
// *                           associated with the following job states.
// *                           state_count[0] - Jobs in transit
// *                           state_count[1] - Jobs in a queue
// *                           state_count[2] - Jobs being held
// *                           state_count[3] - Jobs waiting for file stagein
// *                           state_count[4] - Jobs that are running
// *                           state_count[5] - Jobs that are exiting
// *         max_queuable    : An integer identifying the maximum number of 
// *                           jobs that can be placed in the queue.
// *         max_running     : An integer identifying the maximum number of
// *                           jobs that can be concurrently running from this
// *                           queue.
// *         checkpoint_min  : An integer identifying the minimum number of 
// *                           minutes (in cpu time) between checkpointing a
// *                           job.
// *         kill_delay      : An integer containing the number of seconds 
// *                           between when a SIGTERM is sent to a job until
// *                           a SIGKILL is sent.
// *         max_user_run    : An integer containing the maximum number of 
// *                           jobs from a single user that can be concurrently
// *                           running.
// *         max_group_run   : An integer containing the maximum number of 
// *                           jobs from a group that can be concurrently
// *                           running.
// *         alt_router      : A boolean flag, if TRUE an alternative, site-
// *                           supplied router function should be used to 
// *                           route jobs.
// *         route_held_jobs : A boolean flag, if TRUE jobs that have been 
// *                           placed on hold may be routed from this queue.
// *         route_waiting_jobs
// *                         : A boolean flag, if TRUE jobs that are awaiting
// *                           execution may be routed from this queue. 
// *         route_retry_time: An integer that specifies the time delay between
// *                           route retries.
// *         route_lifetime  : An integer that specifies the maximum amount of
// *                           time that a job may exist in a routing queue.
// *	     route_destinations : An array of strings indicating destinations 
// *                           to which a routing queue may forward a job.
// *         from_route_only : If true, this flag indicates that the queue will
// *                           only accept jobs that are sent to it from a 
// *                           routing queue.
// *         acl_host_enable : This boolean flag indicates that the 
// *                           host access control list must be used to
// *                           identify which hosts can attach to the server.
// *         acl_hosts       : An array of hosts that have permission to 
// *                           attach to the server. The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         acl_group_enable: This boolean flag indicates that the 
// *                           group access control list must be used to
// *                           identify which groups can attach to the server.
// *         acl_groups      : An array of groups that have permission to 
// *                           attach to the server. The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         acl_user_enable : This boolean flag indicates that the 
// *                           user access control list must be used to
// *                           identify which users can attach to the server.
// *         acl_users       : An array of users that have permission to 
// *                           attach to the server. The last string in the 
// *                           array is NULL to indicate the end of the list.
// *         resources_default
// *         resources_max
// *         resources_min
// *         resources_available
// *         resources_assigned
// *                         : These are tables of resources that have
// *                           been defined on the server.  See the 
// *                           PBS_Resource class for information on how
// *                           to extract information from these objects.
// *****************************************************************************
class PBS_Queue : public PBS_Base, public PBS_JobList
{
friend class PBS_Server;

public:
	// *********************************************************************
	// * An enumerated list of possible job states.
	// *********************************************************************
	typedef enum
		{
		TRANSIT=0,
		QUEUED,
		HELD,
		WAITING,
		RUNNING,
		EXITING
		} JobStateIndex;    

protected:

	// *********************************************************************
	// * PBS_Server that owns and manages this PBS_Queue.
	// *********************************************************************
	class PBS_Server & server;

	// *********************************************************************
	// * Common queue attributes.
	// *********************************************************************
	char             * name;
	QueueType          queue_type;
	Boolean            enabled;
	Boolean            started;
	int                priority;
	size_t             total_jobs;
	StateCount         state_count;
	size_t             max_queuable;
	size_t             max_running;
	Boolean            from_route_only;
	Boolean            acl_host_enable;
	char **            acl_hosts;
	Boolean            acl_group_enable;
	char **            acl_groups;
	Boolean            acl_user_enable;
	char **            acl_users;
	PBS_Resource       resources_default;
	PBS_Resource       resources_max;
	PBS_Resource       resources_min;
	
	// *********************************************************************
	// * Execution queue attributes
	// *********************************************************************
	int                checkpoint_min;
	int                kill_delay;
	size_t             max_user_run;
	size_t             max_group_run;
	PBS_Resource       resources_available;
	PBS_Resource       resources_assigned;
	
	
	// *********************************************************************
	// * Routing queue attributes
	// *********************************************************************
	Boolean            alt_router;
	Boolean            route_held_jobs;
	Boolean            route_waiting_jobs;
	int                route_retry_time;
	int                route_lifetime;
	char **            route_destinations;
	
public:
	// *********************************************************************
	// * PBS_Queue::PBS_Queue :
	// *	This constructor is called with a reference to the PBS_Server
	// *	object and a pointer to a batch_status structure that contains
	// *	configuration information for the queue.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************	
	PBS_Queue ( PBS_Server & Server, struct batch_status * Stat )
		: server             (Server),
		  name               (NULL),
		  queue_type         (EXECUTION),
		  enabled            (FALSE),
		  started            (FALSE),
		  priority           (0),
		  total_jobs         (0),
		  max_queuable       (0),
		  max_running        (0),
		  from_route_only    (FALSE),
		  acl_host_enable    (FALSE),
		  acl_hosts          (NULL),
		  acl_group_enable   (FALSE),
		  acl_groups         (NULL),
		  acl_user_enable    (FALSE),
		  acl_users          (NULL),
		  checkpoint_min     (0),
		  kill_delay         (2),
		  max_user_run       (0),
		  max_group_run      (0),
		  alt_router         (FALSE),
		  route_held_jobs    (FALSE),
		  route_waiting_jobs (FALSE),
		  route_retry_time   (30),
		  route_lifetime     (0),
		  route_destinations (NULL),
		  resources_default  (),
		  resources_max      (),
		  resources_min      (),
		  resources_available(),
		  resources_assigned (),
		  PBS_JobList        ()
		{
		loadData( Stat );
		}
				

	// *********************************************************************
	// * PBS_Queue::~PBS_Queue :
	// *	The destructor for the object.  It calls the clear method 
	// *	to release all memory that was allocated during the loadData
	// *    process.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	~PBS_Queue ( void )
		{
		clear();
		}
		
	// *********************************************************************
	// * PBS_Queue::loadData :
	// *	Populates the object using the data retrieved from the
	// *	pbs_statque function. Calls various helper functions to parse 
	// *	and store this information.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	int  loadData   ( struct batch_status * job );
	
	// *********************************************************************
	// * PBS_Queue::clear :
	// *	Frees allocated memory and reinitializes variables to their
	// *	default values.
	// *	
	// *	This method is protected to prevent it's execution by anyone
	// *	except the PBS_Server object.
	// *********************************************************************
	void clear ( void );

public:

	// *********************************************************************
	// * PBS_Queue::asciiDump :
	// *	Generates a text dump of the contents of the object.  This 
	// *	output is particularly useful for debugging and reporting the 
	// *	state of the server.
	// *********************************************************************
	void asciiDump  ( FILE * fp = stdout );

	// *********************************************************************
	// * Accessor methods :
	// *	These methods provide access to the data that is stored inside
	// *	the PBS_Queue object.
	// *********************************************************************
	inline PBS_Server &   getServer             ( void ) { return server; }
	inline const char *   getName               ( void ) { return (const char *)name; }
	inline QueueType      getType               ( void ) { return queue_type; }
	inline Boolean        getEnabledFlag        ( void ) { return enabled; }
	inline Boolean        getStartedFlag        ( void ) { return started; }
	inline int            getPriority           ( void ) { return priority; }
	inline size_t         getTotalJobCount      ( void ) { return total_jobs; }
	inline size_t         getTransitJobCount    ( void ) { return state_count[TRANSIT]; }
	inline size_t         getQueuedJobCount     ( void ) { return state_count[QUEUED]; }
	inline size_t         getHeldJobCount       ( void ) { return state_count[HELD]; }
	inline size_t         getWaitingJobCount    ( void ) { return state_count[WAITING]; }
	inline size_t         getRunningJobCount    ( void ) { return state_count[RUNNING]; }
	inline size_t         getExitingJobCount    ( void ) { return state_count[EXITING]; }
	inline size_t         getMaxQueuable        ( void ) { return max_queuable; }
	inline size_t         getMaxRunning         ( void ) { return max_running; }
	inline size_t         getMaxUserRunning     ( void ) { return max_user_run; }
	inline size_t         getMaxGroupRunning    ( void ) { return max_group_run; }
	inline Boolean        getFromRouteOnlyFlag  ( void ) { return from_route_only; }
	inline Boolean        getHostACLEnabledFlag ( void ) { return acl_host_enable; }
	inline const char **  getHostACL            ( void ) { return (const char **)acl_hosts; }
	inline Boolean        getGroupACLEnabledFlag( void ) { return acl_group_enable; }
	inline const char **  getGroupACL           ( void ) { return (const char **)acl_groups; }
	inline Boolean        getUserACLEnabledFlag ( void ) { return acl_user_enable; }
	inline const char **  getUserACL            ( void ) { return (const char **)acl_users; }
	inline int            getCheckpointMin      ( void ) { return checkpoint_min; }
	inline int            getKillDelay          ( void ) { return kill_delay; }
	inline Boolean        getAltRouterFlag      ( void ) { return alt_router; }
	inline Boolean        getRouteHeldFlag      ( void ) { return route_held_jobs; }
	inline Boolean        getRouteWaitingFlag   ( void ) { return route_waiting_jobs; }
	inline int            getRouteRetryTime     ( void ) { return route_retry_time; }
	inline int            getRouteLifetime      ( void ) { return route_lifetime; }
	inline const char **  getRouteDestinations  ( void ) { return (const char **)route_destinations; }
	inline PBS_Resource * getDefaultResources   ( void ) { return &resources_default; }
	inline PBS_Resource * getAssignedResources  ( void ) { return &resources_assigned; }
	inline PBS_Resource * getAvailableResources ( void ) { return &resources_available; }
	inline PBS_Resource * getMaxResources       ( void ) { return &resources_max; }
	inline PBS_Resource * getMinResources       ( void ) { return &resources_min; }
	
};

// *****************************************************************************
// * class PBS_QueueList:
// *	This is a managed array of PBS_Queue objects.
// *****************************************************************************
class PBS_QueueList 
{
protected:
	PBS_Queue ** queues;
	size_t       queueCnt;
	
	// *********************************************************************
	// * PBS_QueueList::PBS_QueueList :
	// *	Default constructor for the class.  It initializes all internal
	// *	variables to their default values.
	// *********************************************************************
	PBS_QueueList ( void );

	// *********************************************************************
	// * PBS_QueueList::addQueues :
	// *	This method populates the object using an external array of 
	// *	queues that is termimated by a NULL pointer.  The array becomes 
	// *	the property of the class and should not be accessed directly 
	// *	by the caller.
	// *********************************************************************
	void addQueues ( PBS_Queue ** queuelist );
	
public:
	// *********************************************************************
	// * PBS_QueueList::PBS_QueueList :
	// *	Parameterized constructor for the class.  It initializes the
	// *	internal variables using an array of queues that is terminated
	// *	by a NULL pointer.
	// *
	// *	The array becomes the property of the object and should not be
	// *	accessed directly by the caller.
	// *********************************************************************
	PBS_QueueList ( PBS_Queue ** queuelist );
	
	// *********************************************************************
	// * PBS_QueueList::~PBS_QueueList :
	// *	Destructor for the object.  It calls clear to free the array 
	// *	structure associated with the object. The PBS_Server is 
	// *	responsible for deleteing the individual PBS_Queues.
	// *********************************************************************
	~PBS_QueueList ( void );

	// *********************************************************************
	// * PBS_QueueList::getQueue :
	// *	This method returns a specific queue by name.  If the queue doesn't
	// *	exist in the list, then a NULL will be returned.
	// *********************************************************************
	PBS_Queue * getQueue ( char * name );
		
	// *********************************************************************
	// * PBS_QueueList::getQueueCount :
	// *	Returns a count of all queues that are currently in the list.
	// *********************************************************************
	size_t getQueueCount ( void );

	// *********************************************************************
	// * PBS_QueueList::getQueues :
	// *	This method returns an array of all queues that are available in 
	// *	the list.  The user is responsible for deleteing the array 
	// *	that is allocated by this call.
	// *********************************************************************
	PBS_Queue ** getQueues ( void );

	// *********************************************************************
	// * PBS_QueueList::asciiDump :
	// *	Sequentially dumps all of the queues in the list.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout );
	
	// *********************************************************************
	// * PBS_QueueList::operator [] :
	// *	This method allows the caller to walk through the internal array
	// *	of PBS_Queue objects.
	// *********************************************************************
	PBS_Queue * operator [] (size_t idx)
		{
		return (idx<queueCnt)?queues[idx]:NULL;
		}
};
#endif /* _PBS_QUEUE_H_ */
