// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This file realizes the methods for the PBS_ServerDatabase that are 
// 	defined in the header file PBS_ServerDatabase.h.  These objects are
//	stored in a separate library and are typically used by the scheduling
//	system to garner system utilization information for user share 
//	computations.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include "PBS_ServerDatabase.h"

#include <sys/time.h>
#include <sys/types.h>
#include <dirent.h>


#ifdef __linux__ 
	#define DIRENT_SORT_PARM const void *
#else	
	#define DIRENT_SORT_PARM const struct dirent **
#endif

#define PBSEVENT_JOB       0x0008
#define PBSEVENT_JOB_USAGE 0x0010
	
// *****************************************************************************
// * stringToDuration :
// *    This method converts a time string to a time_t. The defaultValue 
// *    provided by the caller will be used in the event that the str is NULL
// *    or blank.
// *****************************************************************************
static time_t stringToDuration ( const char * str, time_t defaultValue )
        {
        time_t result=defaultValue;

        if(str && *str)
                {
                int multx = 1;
                for(const char *ptr=str; multx<3600 && (ptr=strchr(ptr, ':'))!=NULL; ptr++)
                        {
                        multx*=60;
                        }

                result    = 0;
                while(*str && multx)
                        {
                        result += atoi(str)*multx;
                        if(*(str+=strcspn(str, ":"))) str++;
                        multx/=60;
                        }
                }

        return result;
        }

// *****************************************************************************
// * scandirSelect :
// *	This function selects which files found in the dataPath will be read 
// * 	into the PBS_ServerDatabase.  Each PBS log file is named YYYYMMDD to 
// *	reflect the month day and year that it was opened.  This method ensures 
// *	that the filename complies with this format and that it is within a 
// * 	certain time period.
// *
// *	This method restricts the table of data that it manages to 14 days.
// *****************************************************************************
static int scandirSelect (const struct dirent * dir )
	{
	time_t           then      = time(NULL)-(14*24*60*60);
	time_t           fileTime  = 0;
	int              result    = 0;
	struct tm        fileTm;
	
	if(sscanf(dir->d_name, "%04d%02d%02d", &fileTm.tm_year, &fileTm.tm_mon, &fileTm.tm_mday)==3 && !dir->d_name[8])
		{
		fileTm.tm_mon -= 1;
         	fileTm.tm_year-= 1900;
		fileTm.tm_sec  = 59;
		fileTm.tm_min  = 59;
		fileTm.tm_hour = 23;
		fileTime = mktime(&fileTm);
		if(fileTime>then) result = 1;
		}

	return result;
	}
	
// *****************************************************************************
// * scandirCompare :
// *	This method compares the names of two log files for sorting.
// *****************************************************************************
static int scandirCompare ( DIRENT_SORT_PARM a, DIRENT_SORT_PARM b )
	{
	const struct dirent * A = *(const struct dirent **)a;
	const struct dirent * B = *(const struct dirent **)b;
	
	return atoi(B->d_name)-atoi(A->d_name);
	}

// *****************************************************************************
// * PBS_ServerDatabase::PBS_ServerDatabase :
// *	This is the constructor for the PBS_ServerDatabase object.  It will 
// *	initialize objects that are internal to the class and then call
// *	the 'initialize' method to read the log files that reside in the 
// *	DataPath.
// *****************************************************************************
PBS_ServerDatabase::PBS_ServerDatabase ( char * Name, char * DataPath, time_t days )
	: PBS_JobCollection(Name, 256),
	  dataPath(NULL),
	  dataFileOffset(0),
	  cutOffSeconds(days*24*60*60)
	{
	if(DataPath && *DataPath) dataPath=strdup(DataPath);
	else dataPath = strdup("/usr/spool/PBS/server_logs");
	
	*dataFile = 0;
	initialize();
	}

// *****************************************************************************
// * PBS_ServerDatabase::~PBS_ServerDatabase :
// *	This is the destructor for the PBS_ServerDatabase object.  It frees all
// *	memory that was allocated by the class and deletes all of the 
// * 	user/queue PBS_JobArray objects that are stored in the internal hash
// *	tables.
// *****************************************************************************
PBS_ServerDatabase::~PBS_ServerDatabase ( void )
	{	
	if(dataPath) free  (dataPath);

	PBSStringHashIter qIter(queueHash);
	PBSStringHashIter uIter(userHash);
	PBS_JobArray *        qRec;
	PBS_JobArray *        uRec;

	PBS_JobCollection::clear();
	
	while((qRec = (PBS_JobArray *)qIter.data()))
		{
		qIter++;
		delete qRec;
		}
	while((uRec = (PBS_JobArray *)uIter.data()))
		{
		uIter++;
		delete uRec;
		}
	}


// *****************************************************************************
// * PBS_ServerDatabase::newJob :
// *	This method is called to create a new PBS_JobRecord object. Because the
// *	constructor for the PBS_JobRecord object is hidden, this is the only
// *	way to create one.  Using the factory approach insures that a 
// *	valid PBS_ServerDatabase object and job ID is provided to the 
// *	PBS_JobRecord constructor.
// *****************************************************************************
PBS_JobRecord * PBS_ServerDatabase::newJob ( const char * job )
	{
	PBS_JobRecord * jRec;

	if((jRec = getJob(job))==NULL) 
		addJob((jRec = new PBS_JobRecord(*this, job)));

	return jRec;
	}
	
// *****************************************************************************
// * PBS_ServerDatabase::readFile :
// *	This method is called to read a log file that is identified by the 
// *	file pointer (fp).  Each entry that is read is stored in a PBS_JobRecord
// *	entry for future utilization.
// *****************************************************************************
void PBS_ServerDatabase::readFile ( FILE * fp )
	{
	int       error = 0;
	int       eventType = 0;
	char      msgDaemon[11];
	char      objClass [5];
	char      owner  [128];
	char      queue  [128];
	char      job    [128];
	int       status;
	int       cput_hour,  cput_min,  cput_sec;
	int       wallt_hour, wallt_min, wallt_sec;
	char      pmem   [32];
	char      vmem   [32];
	int       dataOffset = 0;
	
	struct tm recTm;
	char      recBuf[4096];
	
	// *********************************************************************
	// * Read a text line from the file.
	// *********************************************************************
	while(fgets(recBuf, 4096, fp)!=NULL)
        	{
		// *************************************************************
		// * If the length of the string is greater than 20, then
		// * the eventType integer can be extracted from the string
		// * starting at recBuf[20]. If this value is PBSEVENT_JOB or
		// * PBSEVENT_JOB_USAGE then the line contains information that
		// * needs to be incorporated into the database.
		// *************************************************************
		if(strlen(recBuf)>20 && (strtol(recBuf+20, NULL, 16)&(PBSEVENT_JOB|PBSEVENT_JOB_USAGE))!=0)
			{
			// *****************************************************
			// * Use sscanf to read the record time, event type,
			// * message daemon, object class and job name from the
			// * record.  At least 10 fields should be read from the
			// * line for this to be a valid record. The %n 
			// * parameter will obtain the offset to the last 
			// * character read from the string.
			// *****************************************************
			if(sscanf( recBuf, 
        		           "%02d/%02d/%04d %02d:%02d:%02d;%04x;%10[^;];%4[^;];%128[^;];%n", 
        		           &recTm.tm_mon, &recTm.tm_mday, &recTm.tm_year,
        		           &recTm.tm_hour, &recTm.tm_min, &recTm.tm_sec,
        		           &eventType, msgDaemon, objClass, job, &dataOffset)>=10)
				{
				// *********************************************
				// * Make the alterations necessary for the 
				// * date to be stored in the proper format.
				// *********************************************
				recTm.tm_mon -=1;
         			recTm.tm_year-=1900;
				
				// *********************************************
				// * Call the newJob method to get a pointer to 
				// * a PBS_JobRecord associated with the jobID read
				// * from the file.  If a corresponding 
				// * PBS_JobRecord already exists, then it will be 
				// * returned - otherwise, a new one will be 
				// * created.
				// *********************************************
				PBS_JobRecord * jrec = newJob(job);
				
				// *********************************************
				// * Set ptr to point to the remaining, 
				// * unprocessed portion of the string...
				// *********************************************
				char      * ptr  = recBuf+dataOffset;
				
				// *********************************************
				// * If the eventType read from the record is
				// * PBSEVENT_JOB, then either a job was run 
				// * or a job was enequeued.  Examine pointer
				// * to determine which of these events occured
				// * and collect the appropriate data from 
				// * the record.
				// *********************************************
				if(eventType&PBSEVENT_JOB)
					{
					// *************************************
					// * If the record is a "Job Run", then
					// * all you can get is the start time.
					// *************************************
					if(!strncmp(ptr, "Job Run", 7))
						jrec->setTimeStarted(mktime(&recTm));
					// *************************************
					// * If the record is a "Job Queued",
					// * then you can get the timeEnqueued,
					// * owner, job name and queue name.
					// *************************************
					else if(!strncmp(ptr, "Job Queued", 10))
						{
						jrec->setTimeEnqueued (mktime(&recTm));
						
						sprintf(owner, "unknown");
						sprintf(queue, "unknown");
						sprintf(job,   "unknown");
						
						sscanf(ptr+strcspn(ptr, ","), 
						       ", owner = %128[^,], job name = %128[^,], queue = %128[^\n]\n", 
						       owner, job, queue);
						
						jrec->setOwner(getUserName(owner));
						jrec->setQueue(queue);
						jrec->setJobName(job);
						}
					}
				// *********************************************
				// * If the eventType is PBSEVENT_JOB_USAGE,
				// * then the job has terminated and the record
				// * contains te job statistics.  Read the 
				// * following data from the record: time
				// * completed, cpu usage, wall time, physical
				// * memory used, and virtual memory used.
				// *********************************************
				else if(eventType&PBSEVENT_JOB_USAGE)
					{
					jrec->setTimeFinished(mktime(&recTm));
					
					status = -1;
					cput_hour  = 0;
					cput_min   = 0;
					cput_sec   = 0;
					*pmem      = 0;
					*vmem      = 0;
					wallt_hour = 0;
					wallt_min  = 0;
					wallt_sec  = 0;
					sscanf(ptr, 
					       "Exit_status=%i resources_used.cput=%04d:%02d:%02d resources_used.mem=%32[^ ] resources_used.vmem=%32[^ ] resources_used.walltime=%04d:%02d:%02d",
					       &status, &cput_hour, &cput_min, &cput_sec, pmem, vmem, &wallt_hour, &wallt_min, &wallt_sec);
					
					long   val;
					char * endPtr;
						
					// *****************************
					// * Convert the physical memory
					// * used to an integer. Don't
					// * forget to consider 'k' and
					// * 'm' for kilobytes or mega
					// * bytes.
					// *****************************
					val = strtol(pmem, &endPtr, 10);
					if(*endPtr=='k') val*=1024;
					else if(*endPtr=='m') val*=1048576;
					jrec->setPhysMemUsed(val);
						
					// *****************************
					// * Convert the virtual memory
					// * used to an integer. Don't
					// * forget to consider 'k' and
					// * 'm' for kilobytes or mega
					// * bytes.
					// *****************************
					val = strtol(vmem, &endPtr, 10);
					if(*endPtr=='k') val*=1024;
					else if(*endPtr=='m') val*=1048576;
					jrec->setVirtMemUsed(val);
						
					jrec->setExitStatus(status);
					jrec->setCPUTimeUsed((long)cput_hour*3600+cput_min*60+cput_sec);
					jrec->setWallTimeUsed((long)wallt_hour*3600+wallt_min*60+wallt_sec);
					}
				}
			}
		}
	// *********************************************************************
	// * Call clearerr to clear the end-of-file marker.  This will allow
	// * the tail of the file to be followed later.
	// *********************************************************************
	clearerr(fp);
	}

// *****************************************************************************
// * PBS_ServerDatabase::initialize :
// *	This method is called by the contstructor to begin loading records
// *	from the log files.  
// *****************************************************************************
void PBS_ServerDatabase::initialize ( void )
	{
	struct dirent **dataFiles = NULL;

	// *********************************************************************
	// * Call  clear to insure there are no lingering jobs.
	// *********************************************************************
	clear();
	
	// *********************************************************************
	// * Initialize the dataFile and dataFileOffset variables.
	// *********************************************************************
	*dataFile      = 0;
	dataFileOffset = 0;
	
	// *********************************************************************
	// * First collect the names of all of the files that are in the 
	// * dataPath directory. Files are included or excluded by the 
	// * scandirSelect function.
	// *********************************************************************
	int filecnt = scandir(dataPath, &dataFiles, scandirSelect, scandirCompare);

	// *********************************************************************
	// * Walk through each of the selected files and call the readFile 
	// * method to assimilate their contents.
	// *********************************************************************
	while(filecnt>0)
		{
		sprintf(dataFile, "%s/%s", dataPath, dataFiles[--filecnt]->d_name);
		
		// *************************************************************
		// * Note: all files will be processed except the most recent
		// * one...  The most recent file will be polled periodically
		// * by PBS_ServerDatabase::update to read new log entries.
		// *************************************************************
		if(filecnt)
			{
			FILE * fp;
			if((fp = fopen(dataFile, "r"))!=NULL)
				{
				readFile(fp);
				fclose  (fp);
				}
			else LogError("PBS_ServerDatabase", "Failed to open %s", dataFile);
			}
		else	{
			dataFileOffset = 0;
			update();
			}
		// *************************************************************
		// * Call free here to release the dirent structure that was
		// * allocated by the scandir function...
		// *************************************************************
		free(dataFiles[filecnt]);
		}
	// *********************************************************************
	// * Call free here to release the array of dirent pointers that was
	// * allocated by the scandir function.
	// *********************************************************************
	free(dataFiles);
	}


// *****************************************************************************
// * PBS_ServerDatabase::update :
// *	This method is called to read any records that have been logged since
// *	initialization or since the last update.  If the date has changed, then
// *	the current accounting file will be closed and the next file will be
// *	opened and processed.
// *****************************************************************************
void PBS_ServerDatabase::update ( void )
	{
	FILE *    fp    = NULL;
	time_t    now   = time(NULL);
	struct tm today = *localtime(&now);
	
	char      newFile[9];
	sprintf(newFile, "%04d%02d%02d", today.tm_year+1900, today.tm_mon+1, today.tm_mday);
	
	if(*dataFile && (fp=fopen(dataFile, "r"))!=NULL)
		{
		if(fseek(fp, dataFileOffset, SEEK_SET)==0)
	  		{
			readFile(fp);
			dataFileOffset = (int)ftell(fp);
			}
		fclose(fp);
		}
	if(!*dataFile || strcmp(newFile, dataFile+strlen(dataFile)-8))
		{
		sprintf(dataFile, "%s/%s", dataPath, newFile);
		dataFileOffset = 0;
		update();
		}
	
	// *********************************************************************
	// * Call purge to remove any jobs that may have aged beyond usefulness
	// *********************************************************************
	purge();
	}	
	
// *****************************************************************************
// * PBS_ServerDatabase::update :
// *	This method is called to refresh the content of running jobs using the
// *	data from the PBS_Server.
// *****************************************************************************
void PBS_ServerDatabase::update ( PBS_Server & server )
	{
	// *********************************************************************
	// * First call the standard update to refresh the database from the
	// * records on disk.
	// *********************************************************************
	update();
	
	// *********************************************************************
	// * Get a list of running jobs from the server.
	// *********************************************************************
	PBS_Job ** running = server.getJobsByState(PBS_Job::RUNNING);
	size_t     idx     = 0; 
	
	// *********************************************************************
	// * Walk through the list of running jobs and update the PBS_JobRecord
	// * object using the current wall and cpu times.
	// *********************************************************************
	for(idx=0; running && running[idx]; idx++)
		{
		PBS_JobRecord * rec = getJob(running[idx]->getName());
		if(rec)
			{
			rec->setWallTimeUsed(stringToDuration(running[idx]->getUsedResources()->find("walltime"), 0));
			rec->setCPUTimeUsed (stringToDuration(running[idx]->getUsedResources()->find("cput"), 0));
			}
		}
	// *********************************************************************
	// * Delete the temporary 'running' array that was returned by the
	// * PBS_Server object --- don't delete the actual objects.
	// *********************************************************************
	delete[] running;
	}	

// *****************************************************************************
// * PBS_ServerDatabase::clear :
// *	This method will clear all of the PBS_JobRecord from the PBS_ServerDatabase.
// *	This approach is generally taken prior to reinitializing from the 
// *	log files.
// *****************************************************************************
void PBS_ServerDatabase::clear ( void )
	{
	PBSStringHashIter qIter(queueHash);
	PBSStringHashIter uIter(userHash);
	PBS_JobArray *    qRec;
	PBS_JobArray *    uRec;

	// *********************************************************************
	// * Clear all PBS_JobRecords associated with the PBS_ServerDatabase.
	// *********************************************************************
	PBS_JobCollection::clear();
	
	// *********************************************************************
	// * Clear all PBS_JobRecords associated with the queues.
	// *********************************************************************
	while((qRec = (PBS_JobArray *)qIter.data()))
		{
		qIter++;
		qRec->clear();
		}
	
	// *********************************************************************
	// * Clear all PBS_JobRecords associated with the users.
	// *********************************************************************
	while((uRec = (PBS_JobArray *)uIter.data()))
		{
		uIter++;
		uRec->clear();
		}
	}
	
// *****************************************************************************
// * PBS_ServerDatabase::purge :
// *	This method will remove all database entries for jobs that completed
// *	prior to 'cutOffSeconds' seconds in the past. By default this value 
// *	is 14 days.
// *
// *	Note: Because all jobs in the queueHash and userHash must first 
// *	exist in the jobHash - by deleting a job from the jobHash, it is
// *	also removed from its associated user/queue PBS_JobArray.
// *****************************************************************************
void PBS_ServerDatabase::purge ( void )
	{
	time_t old = time(NULL)-(cutOffSeconds);
	
	PBSStringHashIter jIter(jobHash);
	PBS_JobRecord *       job;
	
	while((job = (PBS_JobRecord *)jIter.data()))
		{
		jIter++;
		time_t fin = job->getTimeFinished();
		if(fin>0 && fin<old) delete job;
		}
	
	// *********************************************************************
	// * Sort the contents of the user/queue and PBS_JobArrays.
	// *********************************************************************
	PBSStringHashIter qIter(queueHash);
	PBSStringHashIter uIter(userHash);
	PBS_JobArray *        qRec;
	PBS_JobArray *        uRec;

	// *********************************************************************
	// * Clear all PBS_JobRecords associated with the queues.
	// *********************************************************************
	while((qRec = (PBS_JobArray *)qIter.data()))
		{
		qIter++;
		qRec->sort();
		}
	
	// *********************************************************************
	// * Clear all PBS_JobRecords associated with the users.
	// *********************************************************************
	while((uRec = (PBS_JobArray *)uIter.data()))
		{
		uIter++;
		uRec->sort();
		}
	}		
	
// *****************************************************************************
// * PBS_ServerDatabase::asciiDump :
// *	This method generates a diagnostic dump of the contents of the 
// *	PBS_ServerDatabase object.  This method is useful for debugging or for
// *	general purpose output.
// *****************************************************************************
void PBS_ServerDatabase::asciiDump ( FILE * fp , char * prefix )
	{
	char blank = '\0';
	prefix = (prefix?prefix:&blank);
	char * newPrefix = new char [strlen(prefix)+6];
	sprintf(newPrefix, "%s     ", prefix);
	
	fprintf(fp, "%sContents of Server Database for %s\n", prefix, name);
	
	fprintf(fp, "\n%sJOB DATA\n%s----------\n\n", newPrefix, newPrefix);
	PBS_JobCollection::asciiDump(fp, newPrefix);

	fprintf(fp, "\n%sUSER DATA\n%s----------\n", newPrefix, newPrefix);
	PBSStringHashIter uIter(userHash);
	PBS_JobArray * uRec;
	while((uRec = (PBS_JobArray *)uIter.data()))
		{
		uRec->asciiDump(fp, newPrefix);
		uIter++;
		}

	fprintf(fp, "\n%sQUEUE DATA\n%s----------\n", newPrefix, newPrefix);
	PBSStringHashIter qIter(queueHash);
	PBS_JobArray * qRec;
	while((qRec = (PBS_JobArray *)qIter.data()))
		{
		qRec->asciiDump(fp, newPrefix);
		qIter++;
		}
	
	delete newPrefix;
	}
	
// *****************************************************************************
// * PBS_ServerDatabase::getUserName :
// *	This method will extact the user name portion of a user@host designation
// *	making it possible for users to be tracked by local name only. The user
// *	name provided here is a static variable and should only be user in a 
// *	transient fashion by the caller.
// *****************************************************************************
char * PBS_ServerDatabase::getUserName ( const char * name )
	{
	static char localUser[256];
	size_t idx = 0;
	for(idx=0; idx<255 && name && name[idx] && name[idx]!='@'; idx++);
	if(idx>0) 
		{
		strncpy(localUser, name, idx);
		localUser[idx] = 0;
		}
	else	{
		strcpy(localUser, "unknown");
		}
	return localUser;
	}
	
// *****************************************************************************
// * PBS_ServerDatabase::getQueue :
// *	This method allows the caller to obtain a PBS_JobArray pointer that
// *	is associated with the user provided queue name.  If a record doesn't
// *	exist for the specified queue, then a new one will be created and 
// *	returned.  The newly created PBS_JobArray remains the property of the 
// *	PBS_ServerDatabase.
// *****************************************************************************
PBS_JobArray * PBS_ServerDatabase::getQueue ( const char * Queue )
	{
	PBS_JobArray * queue = NULL;
	
	if(Queue && *Queue && (queue = (PBS_JobArray *)queueHash.find(Queue))==NULL)
		{
		queue = new PBS_JobArray(Queue, 32);
		queueHash.insert((char *)queue->getName(), queue);
		}

	return queue;
	}
	
// *****************************************************************************
// * PBS_ServerDatabase::getUser :
// *	This method allows the caller to obtain a PBS_JobArray pointer that
// *	is associated with the user provided name.  If a record doesn't
// *	exist for the specified user name, then a new one will be created and
// *	returned.  The PBS_JobArray object remains the property of the 
// *	PBS_ServerDatabase.
// *****************************************************************************
PBS_JobArray * PBS_ServerDatabase::getUser ( const char * User )
	{
	PBS_JobArray * user = NULL;

	if(User && *User && (user = (PBS_JobArray *)userHash.find(getUserName(User)))==NULL)
		{
		user = new PBS_JobArray(getUserName(User), 32);
		userHash.insert((char *)user->getName(), user);
		}
	
	return user;
	}
	
