// -----------------------------------------------------------------------------
//  Copyright (c) 2001     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <PBS_Server.h>
#include <PBS_Accounting.h>
#include <unistd.h>

// *****************************************************************************
// * This declaratation is required by the Logging Daemon within the PBS
// * library.
// *****************************************************************************
const char * msg_daemonname = SetLogLocal("usage.cc");

// *****************************************************************************
// * struct UsageEvent :
// *	This structure contains a time and an integer that reflects the current
// *	system utilization at that time (from 0.0 - 1.0 ).
// *****************************************************************************
typedef struct
	{
	time_t moment;
	double usage;
	} UsageEvent;
	

// *****************************************************************************
// * struct NodeDef :
// *	This structure defines a node name and the number of processors that it
// *	contains.
// *****************************************************************************
typedef struct
	{
	char * name;
	size_t procCnt;
	} NodeDef;
	
// *****************************************************************************
// * UsageEventSort :
// *	This is a static function that is used by qsort to sort an array of 
// *	UsageEvent objects based on the time that they occurred.
// *****************************************************************************
static int UsageEventSort (const void * ptr1, const void * ptr2 )
	{
	UsageEvent & event1 = *(UsageEvent *)ptr1;
	UsageEvent & event2 = *(UsageEvent *)ptr2;

	return (event1.moment - event2.moment);
	}

// *****************************************************************************
// * max and min :
// *	These are inline versions of the popular max and min definitions 
// *	defined within the 'C' language.
// *****************************************************************************
inline time_t max ( time_t a, time_t b ) { return (a>b)?a:b; }
inline time_t min ( time_t a, time_t b ) { return (a<b)?a:b; }


// *****************************************************************************
// * composeEvents :
// *	This method will walk through all of the events in the accounting 
// *	database and will create a series of UsageEvent objects that represent
// *	the combined system utilization on the specified nodes. This method will'
// *	return the total number of events that were extracted from the database.
// *
// *	acct :		Is a fully populated PBS Accounting Database object.
// *	nodes :		Is an array of NodeDef objects that identify the nodes 
// *			that should be evaluated.
// *	nodeCnt :	Is the number of NodeDef objects.
// *	events :	Is the UsageEvent object that will be populated with
// *			statistics about the specified nodes' combined usage.
// *****************************************************************************
size_t composeEvents (	PBS_AccountingDatabase & acct,
			NodeDef                  nodes[],
			size_t                   nodeCnt,
			UsageEvent            *& events )
	{
	size_t idx;
	size_t procTotal = 0;
	size_t recTotal  = 0;
	size_t recIdx    = 0;
	
	// *********************************************************************
	// * First add up all of the processors in all of the nodes. At the same
	// * time count the number of records that exist for each of the nodes.
	// *
	// * Remember that since each record represents two events (a start
	// * and an end, the recTotal will need to be multiplied by two.
	// *********************************************************************
	for(idx=0; idx<nodeCnt; idx++)
		{
		PBS_NodeAccounting * nAcct  = acct.findNode(nodes[idx].name);
		recTotal += nAcct?(nAcct->getCount()*2):0;
		procTotal+=nodes[idx].procCnt;
		}

	// *********************************************************************
	// * If the total number of records is greater than 0, allocate a 
	// * sufficient number of UsageEvent objects to hold them --- otherwise
	// * return 0.
	// *********************************************************************
	if(recTotal) events = (UsageEvent *)malloc(recTotal*sizeof(UsageEvent));
	else return 0;
	
	// *********************************************************************
	// * Accrue all of the events from the accounting database into the
	// * UsageEvent objects.  At this time we are merely going to set the
	// * usage value to the number of processors employed by the event.
	// *
	// * A start event that uses two processors will have a usage of +2.0
	// * while the associated end event will have a usage of -2.0
	// *********************************************************************
	for(idx=0; idx<nodeCnt; idx++)
		{
		PBS_NodeAccounting * nAcct = acct.findNode(nodes[idx].name);
		if(nAcct)
			{
			PBS_AccountingRecord ** recs = nAcct->getRecords();
			size_t                  nCnt = nAcct->getCount();
			size_t                  nIdx = 0;
			
			for(nIdx=0; nIdx<nCnt; nIdx++)
				{
				int procs = recs[nIdx]->getNodeProcCnt(nodes[idx].name);
				if(procs)
					{
					events[recIdx].moment  = recs[nIdx]->getStartTime();
					events[recIdx++].usage = procs;
					
					// *************************************
					// * Note: If the job IS RUNNING then
					// * it will have an end time of zero 
					// * and should not be recorded.
					// *************************************
					if(recs[nIdx]->getEndTime())
						{
						events[recIdx].moment  = recs[nIdx]->getEndTime();
						events[recIdx++].usage = -procs;
						}
					}
				}
			}
		}
	
	// *********************************************************************
	// * Once all of the events have been accrued, sort them in time order.
	// * Note that since 0 proc events were discared, that the value of
	// * recTotal will now be decreased to the current value of recIdx;
	// *********************************************************************
	recTotal = recIdx;
	qsort(events, recTotal, sizeof(UsageEvent), UsageEventSort);

	// *********************************************************************
	// * Now that all of the events are time ordered, combine all of the 
	// * events that occur at the same time. This will also elliminate any
	// * combination that resulted in a zero event.  Again, because of 
	// * record ellimination, the recTotal value will be reduced to recIdx.
	// *********************************************************************
	for(idx=0, recIdx=0; idx<recTotal; )
		{
		if(idx!=recIdx)
			{
			events[recIdx].moment = events[idx].moment;
			events[recIdx].usage  = events[idx].usage;
			}
		for(	idx++; 
			idx<recTotal && events[recIdx].moment==events[idx].moment; 
			idx++)
			{
			events[recIdx].usage += events[idx].usage;
			}
		if(events[recIdx].usage!=0) recIdx++;
		}
	recTotal = recIdx;

	// *********************************************************************
	// * Next, walk through the list and add each usage value to the 
	// * previous usage value.  This will give a running tally of the number
	// * of processors in use at each time interval.
	// *********************************************************************
	for(idx=1; idx<recTotal; idx++)
		{
		events[idx].usage += events[idx-1].usage;
		}
	
	// *********************************************************************
	// * Finally, divide the number of processors in use at each time
	// * interval by the total number of processors available to create
	// * a utilization percentage.
	// *********************************************************************
	for(idx=0; idx<recTotal; idx++)
		{
		events[idx].usage /= (double)procTotal;
		}
	
	// *********************************************************************
	// * Reallocate the events object to free unneeded memory.
	// *********************************************************************
	events = (UsageEvent *)realloc(events, recTotal*sizeof(UsageEvent));
	
	return recTotal;
	}
	

// *****************************************************************************
// * getPeriodEvents :
// *	This method will extract a list of UsageEvents that occurred within
// *	a specified time window. If no events are specified within the events
// *	object (or if all events happen outside of range), then one object will 
// *	be created at the start time with a usage of zero.
// *****************************************************************************
size_t getPeriodEvents ( UsageEvent *  events, 
			 size_t        eventCnt,
			 time_t        start, 
			 time_t        end, 
			 UsageEvent *& output )
	{
	size_t sIdx;
	size_t eIdx;
	size_t totalRecs  = 0;
	double firstUsage = 0.0;
	double lastUsage  = 0.0;
		
	if(end<start) end = start;
	
	for(sIdx = 0; sIdx<eventCnt && events[sIdx].moment<=start; sIdx++)
		firstUsage = events[sIdx].usage;
		
	for(eIdx = sIdx; eIdx<eventCnt && events[eIdx].moment<=end; eIdx++)
		lastUsage = events[eIdx].usage;
		
	output = (UsageEvent *)malloc((1+eIdx-sIdx)*sizeof(UsageEvent));
	
	output[totalRecs].moment   = start;
	output[totalRecs++].usage  = firstUsage;

	while(sIdx<eIdx)
		{
		output[totalRecs].moment  = events[sIdx].moment;
		output[totalRecs++].usage = events[sIdx++].usage;
		}	
	
	return totalRecs;
	}


// *****************************************************************************
// * timeAverageEvents :
// *	This method will take a list of UsageEvents representing the percentage
// *	of utilization at a specified time and will restructure them to be
// *	distributed evenly over time.
// *
// *	events :	Is the UsageEvent array object that is populated with
// *			an ordered list of time/utilization pairs. 
// *	eventCnt :	Is the number of UsageEvent objects in the events array
// *	interval :	Is the time interval (in seconds) for the output object.
// *	output :	Is a UsageEvent array that will be populated with 
// *			time distributed utilization information.
// *****************************************************************************
size_t getTimeAveragedEvents (	UsageEvent  * events,
				size_t        eventCnt,
				time_t        start,
				time_t        end,
				time_t        interval,
				UsageEvent *& result )
	{
	if(end<start)       end      = start;
	if(interval<60)     interval = 60; 
	if((interval%2)!=0) interval++;

	UsageEvent * boundedEvents   = NULL;
	size_t       boundedEventCnt = 0;
	size_t       resultCnt       = 0;
	size_t       idx             = 0;
	size_t       bIdx            = 0;
	size_t       halfInterval    = interval/2;

	boundedEventCnt = getPeriodEvents(events, eventCnt, start-halfInterval, end+halfInterval, boundedEvents);
	
	resultCnt = ((end-start)/interval)+1;
	result = (UsageEvent *)malloc(resultCnt*sizeof(UsageEvent));
	
	for(idx=0; idx<resultCnt; idx++)
		{
		result[idx].moment = start;
		result[idx].usage  = 0; 
		start+=interval;
		}
	
	time_t lastMoment = 0;
	double lastUsage  = 0;
	
	for(idx=0, bIdx=0; idx<resultCnt; )
		{
		time_t nextMoment = (bIdx<boundedEventCnt)?
		                     boundedEvents[bIdx].moment:
				     (result[idx].moment+halfInterval);
				     
		// *************************************************************
		// * If the usage recorded at lastMoment exists within the 
		// * processing interval of this time averaged region --- then
		// * add it (proportionately) to the region.
		// *************************************************************
		if(nextMoment > result[idx].moment-halfInterval)
			{
			time_t rStart = max(result[idx].moment-halfInterval, lastMoment);
			time_t rEnd   = min(result[idx].moment+halfInterval, nextMoment);
			result[idx].usage += lastUsage*(rEnd-rStart)/interval;
			}
			
		// *************************************************************
		// * If the current bounded event occurs within the scope of 
		// * this time averaged region... increment bIdx, so that it 
		// * will be considered on the next pass. Otherwise, simply
		// * increment idx and move to the next time averaged region.
		// *************************************************************
		if ( bIdx<boundedEventCnt &&
		     boundedEvents[bIdx].moment < result[idx].moment+halfInterval)
		     	{
			lastMoment = boundedEvents[bIdx].moment;
			lastUsage  = boundedEvents[bIdx].usage;
			bIdx++;
			}
		else	idx++;
		}

	// *********************************************************************
	// * Free the memory temporarily allocated to the boundedEvents array.
	// *********************************************************************
	free ( boundedEvents);

	return resultCnt;
	}


// *****************************************************************************
// * getArgs :
// *	This method will load the user arguments from the command line.  If 
// *	an argument is not specified, then a default value will be used.
// *
// *	The following defaults will be used:
// *	    serverName : The value in /usr/spool/PBS/server_name or 
// *	                 the local host name.
// *	    nodes      : The list of all nodes obtained from the server.
// *	    start      : 30 days prior to the end time
// *	    end        : now
// *	    interval   : 1 day
// *
// *	In the event that one of more of the parameters is improperly specified,
// *	the method will report the error and the correct syntax for specifying
// *	arguments and will return -1 --- indicating that no further processing
// *	is possible.
// *****************************************************************************
int getArgs (	int        argc, 
		char    ** argv, 
		char    *& serverName, 
		char    *& accountingPath,
		NodeDef *& nodes, 
		size_t   & nodeCnt, 
		time_t   & start, 
		time_t   & end,
		time_t   & interval,
		int      & showNodesFlag,
		int      & showTotalFlag )
	{
	serverName     = NULL;
	accountingPath = NULL;
	nodes          = NULL;
	nodeCnt        = 0;
	end            = 0;
	start          = 0;
	interval       = 24*60*60;
	showNodesFlag  = 1;
	showTotalFlag  = 1;
	
	int showHelp   = 0;
	int dryRun     = 0;
	int result     = 0;
	int idx        = 0;
	int connID     = 0;
	
	// *********************************************************************
	// * Process each of the arguments provided to the application.
	// *********************************************************************
	for(idx=1; idx<argc; idx++)
		{
		if(*argv[idx] == '-') switch(*(argv[idx]+1))
			{
			// *****************************************************
			// * option -T : Report only cumulative totals
			// *****************************************************
			case 'T':
			showTotalFlag = 1;
			showNodesFlag = 0;
			break;
			
			// *****************************************************
			// * option -N : Report only node level averages
			// *****************************************************
			case 'N':
			showNodesFlag = 1;
			showTotalFlag = 0;
			break;
			
			// *****************************************************
			// * option -A : Report both cumulative and node averages
			// *****************************************************
			case 'A':
			showNodesFlag = 1;
			showTotalFlag = 1;
			break;
			
			// *****************************************************
			// * option -s serverName : Specify the name of the
			// * PBS server.
			// *****************************************************
			case 's':
			if(idx+1 < argc && *argv[idx+1] != '-' && !serverName)
				{
				serverName = strdup(argv[++idx]);
				}
			break;
			
			// *****************************************************
			// * option -d path : Specify the path of the accounting
			// * files.
			// *****************************************************
			case 'd':
			if(idx+1 < argc && *argv[idx+1] != '-' && !accountingPath)
				{
				accountingPath = strdup(argv[++idx]);
				}
			break;
			
			// *****************************************************
			// * option -i interval : Specify the reporting interval
			// * in minutes.
			// *****************************************************
			case 'i':
			if(idx+1 < argc && isdigit(*argv[idx+1]))
				{
				interval = atoi(argv[++idx])*60;
				}
			break;
			
			// *****************************************************
			// * option -S "mm/dd/yyyy hh:mm" : Specify the start
			// * time.
			// *****************************************************
			case 'S':
			if(idx+1 < argc && isdigit(*argv[idx+1]))
				{
				time_t t        = time(NULL);
				struct tm tmNow;
				
				memmove(&tmNow, localtime(&t), sizeof(struct tm));
				switch(sscanf(argv[++idx], 
				       		"%d/%d/%d %d:%d",
				       		&tmNow.tm_mon,
				       		&tmNow.tm_mday,
						&tmNow.tm_year,
						&tmNow.tm_hour,
						&tmNow.tm_min))
					{
					case 5:
					case 4:
					case 3:
					if(tmNow.tm_year > 1900) tmNow.tm_year-=1900;
					else if(tmNow.tm_year < 69 ) tmNow.tm_year+=100;

					case 2:
					case 1:
					tmNow.tm_mon -= 1;
					start = mktime(&tmNow);
					break;
					
					default:
					start = 0;
					break;					
					}

				}
			break;

			// *****************************************************
			// * option -E "mm/dd/yyyy hh:mm" : Specify the end time
			// *****************************************************
			case 'E':
			if(idx+1 < argc && isdigit(*argv[idx+1]))
				{
				time_t t        = time(NULL);
				struct tm tmNow;
				
				memmove(&tmNow, localtime(&t), sizeof(struct tm));
				switch(sscanf(argv[++idx], 
				       		"%d/%d/%d %d:%d",
				       		&tmNow.tm_mon,
				       		&tmNow.tm_mday,
						&tmNow.tm_year,
						&tmNow.tm_hour,
						&tmNow.tm_min))
					{
					case 5:
					case 4:
					case 3:
					if(tmNow.tm_year > 1900) tmNow.tm_year-=1900;
					else if(tmNow.tm_year < 69 ) tmNow.tm_year+=100;

					case 2:
					case 1:
					tmNow.tm_mon -= 1;
					end = mktime(&tmNow);
					break;
					
					default:
					end = 0;
					break;					
					}

				}
			break;
			
			// *****************************************************
			// * option -n node1:procs+node2:procs : Specify the 
			// * nodes that should be reported on.
			// *****************************************************
			case 'n':
			if(idx+1 < argc && *argv[idx+1]!='-')
				{
				char * ptr    = argv[++idx];
				char * sPtr   = NULL;
				char   buf[1024];

				nodeCnt = 1;
				while((ptr = strchr(ptr, '+'))!=NULL) 
					{
					ptr++;
					nodeCnt++;
					}
				nodes = new NodeDef[nodeCnt];
				
				ptr     = argv[idx];
				nodeCnt = 0;
				while( ptr && *ptr )
					{
					sPtr = ptr+strcspn(ptr, "+:");
					strncpy(buf, ptr, sPtr-ptr);
					buf[sPtr-ptr] = 0;

					if(*buf)
						{
						nodes[nodeCnt].name = strdup(buf);
						if(*sPtr == ':')
							{
							nodes[nodeCnt++].procCnt = atoi(++sPtr);
							sPtr = sPtr+strcspn(sPtr, "+");
							}
						else nodes[nodeCnt++].procCnt = 0;
						}
					ptr = (sPtr && *sPtr)?++sPtr:sPtr;
					}
				if(nodeCnt==0)
					{
					free(nodes);
					nodes=NULL;
					}
				}
			break;
			
			case 'D':
			dryRun = 1;
			break;
			
			case 'h':
			showHelp = 1;
			break;
			}
		}

	// *********************************************************************
	// * At this point, set the default value for any parameter that was
	// * not specified.
	// *********************************************************************

	// *********************************************************************
	// * If the server name was not specified, gather it from the file
	// * /usr/spool/PBS/server_name.  If this file does not exist, then 
	// * use the local host name.
	// *********************************************************************
	if(!result && serverName==NULL)
		{
		FILE * fp  = NULL;
		char * ptr = NULL;
		char   buf[1024];

		if ((fp = fopen("/usr/spool/PBS/server_name", "r"))!=NULL &&
		    (fgets(buf, 1024, fp)!=NULL))
			{
			// Remove the trailing carriage return
			if((ptr = strchr(buf, '\n'))!=NULL) *ptr = 0;
			}
		else gethostname(buf, 1024);

		serverName = strdup(buf);
		
		}

	// *********************************************************************
	// * Connect to the server to gather statistics about the nodes.
	// *********************************************************************
	if(!result && (connID=pbs_connect(serverName))>0)
		{		
		PBS_Server     server(connID);

		// *************************************************************
		// * If no nodes were specified, then collect all of the known
		// * nodes from the server.
		// *************************************************************
		if(nodeCnt==0)
			{
			PBS_NodeList & nodeList = server;
			for(nodeCnt=0; nodeList[nodeCnt]!=NULL; nodeCnt++);
			if(nodeCnt==0)
				{
				fprintf(stderr, "Error: There are no nodes specified in server %s\n", serverName);
				result = -1;
				}
			else	{
				nodes = new NodeDef[nodeCnt];
				for(idx=0; idx<nodeCnt; idx++)
					{
					nodes[idx].name = strdup(nodeList[idx]->getName());
					nodes[idx].procCnt = nodeList[idx]->getNumProcessors();
					}
				}
			}
		// *************************************************************
		// * Otherwise, obtain the number of processors for each node
		// * if the value was not otherwise specified by the caller.
		// *************************************************************
		else for(idx=0; idx<nodeCnt; idx++)
			{
			if(nodes[idx].procCnt==0)
				{
				PBS_Node * node = server.getNode(nodes[idx].name);
				if(node) nodes[idx].procCnt = node->getNumProcessors();
				else	{
					fprintf(stderr, 
						"Warning: Node %s is not registered on server %s\n", 
						nodes[idx].name, serverName);
					nodes[idx].procCnt = 1;
					}
				}
			}
		}
	// *********************************************************************
	// * If the server could not be contacted, then set the number of 
	// * processors on all of the specified nodes to 1 if not otherwise
	// * specified.
	// *********************************************************************
	else if(!result)
		{
		if(nodeCnt==0)
			{
			fprintf(stderr, "Error: Could not connect to server %s\n", serverName); 
			result = -1;
			}
		else	{
			fprintf(stderr, "Warning: Could not connect to server %s\n", serverName);
			for(idx=0; idx<nodeCnt; idx++)
				{
				if(nodes[idx].procCnt == 0) nodes[idx].procCnt = 1;
				}
			}				
		}

	// *********************************************************************
	// * If the end time was not specified, then use the current time.
	// *********************************************************************
	if(!result && end==0) end = time(NULL);
	
	// *********************************************************************
	// * If the start time was not specified (or if it is greater than the
	// * end time), then use the 30 days before the specified end time.
	// * Also, normalize the start time to begin on a 1 hour boundary.
	// *********************************************************************
	if(!result && (start==0 || start>end ))
		{
		start  = end-(30*24*60*60);
		start -= start%3600; 
		}	
	
	// *********************************************************************
	// * If the reporting interval was not specified - or the reporting 
	// * interval is less than 1 minute - set the reporting interval to 1
	// * hour.
	// *********************************************************************
	if(!result && interval<60) interval = 3600;
	
	// *********************************************************************
	// * Finally, ensure that the parameters are requesting something...
	// * either a cumulative result of a node wise result.
	// *********************************************************************
	if(!result && !showNodesFlag && !showTotalFlag) result = -1;
	
	// *********************************************************************
	// * If the dryRun flag was set... simply report the parameters that
	// * were loaded and return -1 without performing any computations.
	// *********************************************************************
	if(!result && dryRun)
		{
		int longestName = 0;
		
		for(int idx=0; idx<nodeCnt; idx++)
			{
			int nameLen = strlen(nodes[idx].name);
			if(nameLen>longestName) longestName=nameLen;
			}
			
		fprintf(stdout, "=========================================================\n");
		fprintf(stdout, "Dry run settings for utilization: \n");
		fprintf(stdout, " - PBS Server Name : %s\n", serverName);
		fprintf(stdout, " - Database Path   : %s\n", accountingPath?accountingPath:"default");
		fprintf(stdout, " - Start Time      : %s", ctime(&start));
		fprintf(stdout, " - End Time        : %s", ctime(&end));
		fprintf(stdout, " - Interval        : %d minutes\n", interval/60); 
		fprintf(stdout, " - Days of History : %d\n", (time(NULL)-(start-(interval/2)))/(3600*24)+1);
		fprintf(stdout, " - Compute Nodes   : %s\n", showNodesFlag?"yes":"no");
		fprintf(stdout, " - Compute Totals  : %s\n", showTotalFlag?"yes":"no");
		fprintf(stdout, " - PBS Nodes       : ");

		for(int idx=0; idx<nodeCnt; idx++)
			{
			fprintf(stdout, "%-*s [%i processors]\n%21c", 
				longestName, nodes[idx].name, nodes[idx].procCnt, 32);
			}
		fprintf(stdout, "\n=========================================================\n");

		result = 1;
		}
	
	// *********************************************************************
	// * If the user had an error -or- requested the online help
	// *********************************************************************
	if(result==-1 || showHelp)
		{
		fprintf(stdout, "=========================================================\n");
		fprintf(stdout, "usage is : utilization -T | -N | -A\n");
		fprintf(stdout, "                       -s [PBS server name]\n");
		fprintf(stdout, "                       -d [path to accounting files]\n");
		fprintf(stdout, "                       -i [reporting interval]\n");
		fprintf(stdout, "                       -S [start time]\n");
		fprintf(stdout, "                       -E [end time]\n");
		fprintf(stdout, "                       -D\n");
		fprintf(stdout, "                       -h\n");
		fprintf(stdout, "                       -n [node list]\n\n");
		fprintf(stdout, "Options:\n");
		fprintf(stdout, "    -T : compute cumulative averages only\n");
		fprintf(stdout, "    -N : compute averages for each node only\n");
		fprintf(stdout, "    -A : compute both cumulative and node averages\n");
		fprintf(stdout, "    -s : specify the name of the PBS server\n");
		fprintf(stdout, "    -d : specify location of the PBS accounting files\n");
		fprintf(stdout, "    -i : reporting interval (in minutes)\n");
		fprintf(stdout, "    -S : start time specified as \"mm/dd/yyyy hh:mm\"\n");
		fprintf(stdout, "    -E : end time specified as \"mm/dd/yyyy hh:mm\"\n");
		fprintf(stdout, "    -D : dry run - display parameters and quit\n");
		fprintf(stdout, "    -h : display help\n");
		fprintf(stdout, "    -n : list of nodes specified as \n");
		fprintf(stdout, "         node1:procs+node2:procs+node3:procs\n\n");
		fprintf(stdout, "         If no nodes are specified, then all nodes\n");
		fprintf(stdout, "         registered on the server will be used. If the\n");
		fprintf(stdout, "         number of processors is not specified, then the\n");
		fprintf(stdout, "         PBS server will be queried for those values.\n");
		fprintf(stdout, "=========================================================\n");
		
		result = 1;
		}

	return result;
	}


int main ( int argc, char ** argv)
	{
	char    * serverName     = NULL;
	char    * accountingPath = NULL;
	NodeDef * nodes          = NULL;
	size_t    nodeCnt        = 0;
	time_t    start          = 0;
	time_t    end            = 0;
	time_t    interval       = 0;
	int       showNodesFlag  = 0;
	int       showTotalFlag  = 0;
		
	// *********************************************************************
	// * Parse the arguments provided to the program to ensure that 
	// * sufficient information is available.
	// *********************************************************************
	if(getArgs(argc, argv, serverName, accountingPath, nodes, nodeCnt, start, 
		   end, interval, showNodesFlag, showTotalFlag)==0)
		{
		time_t        days         = (time(NULL)-(start-(interval/2)))/(3600*24)+1;
		size_t        longestName  = 9;
		int           idx          = 0;
		size_t        columnCnt    = (showNodesFlag?nodeCnt:0)+(showTotalFlag?1:0);
		size_t        masterIdx    = (showNodesFlag && showTotalFlag)?nodeCnt:0;
		char       ** labels       = new char * [columnCnt];
		UsageEvent *  rawEvents    = NULL;
		size_t        rawEventCnt  = 0;
		UsageEvent ** avgEvents    = new UsageEvent * [columnCnt];
		size_t     *  avgEventCnt  = new size_t       [columnCnt];

		// *************************************************************
		// * Load accounting data sufficient to cover the number of days
		// * in question.
		// *************************************************************
		PBS_AccountingDatabase acct(serverName, days, 200, accountingPath);
		
		// *************************************************************
		// * Clear the data structures allocated to store the average
		// * values.
		// *************************************************************
		memset(avgEvents,   0, columnCnt*sizeof(avgEvents));
		memset(avgEventCnt, 0, columnCnt*sizeof(size_t));
				
		// *************************************************************
		// * If the user has specified that he is interested in seeing
		// * the statistics for each of the nodes - then compute those
		// * statistics and store them in the avgEvents array.
		// *************************************************************
		if(showNodesFlag)
			{
			for(idx=0; idx<nodeCnt; idx++)
				{
				labels[idx] = nodes[idx].name;
				rawEventCnt = 
					composeEvents(acct, nodes+idx, 1, rawEvents);
						
				avgEventCnt[idx] = 
					getTimeAveragedEvents(
						rawEvents, rawEventCnt, start, end, 
						interval, avgEvents[idx]);

				free(rawEvents);
				rawEvents   = NULL;
				rawEventCnt = 0;
				}
			}

		// *************************************************************
		// * If the user has specified that he is interested in the 
		// * totals for all of the specified nodes - then compute those
		// * statistics and store them at the end of the avgEvents array
		// *************************************************************
		if(showTotalFlag)
			{
			labels[masterIdx] = "Average";
			rawEventCnt = 
				composeEvents( acct, nodes, nodeCnt, rawEvents );
					
			avgEventCnt[masterIdx] = 
				getTimeAveragedEvents(
					rawEvents, rawEventCnt, start, end, 
					interval, avgEvents[masterIdx]);
					
			free(rawEvents);
			rawEvents   = NULL;
			rawEventCnt = 0;
			}

		// *************************************************************
		// * Compute the length of the longest name label that will be
		// * used.  This value will be used to ensure that the table is
		// * evenly distributed.
		// *************************************************************
		for(idx=0; idx<columnCnt; idx++) 
			{
			int x = strlen(labels[idx]);
			if(x>longestName) longestName = x;
			}
							
		// *************************************************************
		// * Write the column headings out.
		// *************************************************************
		fprintf(stdout, "%16s", "Time");
		for(idx=0; idx<columnCnt; idx++) fprintf(stdout, "%*s", longestName, labels[idx]);
		fprintf(stdout, "\n");
		
		// *************************************************************
		// * Write each row of entries out.
		// *************************************************************
		for(idx=0; idx<avgEventCnt[masterIdx]; idx++)
			{
			char buf[20];
			struct tm * t = localtime(&avgEvents[masterIdx][idx].moment);
			strftime(buf, 20, "%m/%d/%Y %H:%M", t);
			fprintf(stdout, "%16s", buf);

			for(size_t col=0; col < columnCnt; col++)
				{
				if(avgEvents[col]!=NULL && idx<avgEventCnt[col])
					{
					fprintf(stdout, "%*.2f%%", longestName-1, avgEvents[col][idx].usage*100.0);
					}
				}
			fprintf(stdout, "\n");
			}
		
		// *************************************************************
		// * Delete any memory that was allocated for accruing data.
		// *************************************************************

		for(idx=0; idx<nodeCnt; idx++)   free(nodes[idx].name);
		for(idx=0; idx<columnCnt; idx++) free(avgEvents[idx]);
		free ( serverName );
		
		delete [] labels;
		delete [] nodes;
		delete [] avgEvents;
		delete [] avgEventCnt;			
		}
	}
	

