// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The file Brokerage.h contains the class definition for the NodeBroker
// 	and NodeCommodity class.  
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _BROKERAGE_H_
#define _BROKERAGE_H_ 1

#include <stdlib.h>		// Needed for size_t
#include <stdio.h>		// Needed for FILE structure and stdout
#include <SchedStatic.h>	// Needed for MAXSUBSETS enumeration

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class ClusterNode;
class NodeBroker;

// *****************************************************************************
// * class NodeCommodity :
// *	This object represents one node and all of the capabilities that are
// *	associated with it.  These objects are stored within the NodeBroker.
// *	These objects are exchanged between the broker and client NodeBroker
// *	objects in an attempt to resolve the jobs requirements.
// *****************************************************************************
class NodeCommodity 
{
protected:
	// *********************************************************************
	// * static NodeCommodity * freeList :
	// *	Since these objects are created and deleted rapidly.  A freeList
	// *	is used to hodl the deleted nodeCommodity objects until they are
	// *	needed again.  When a new nodeCommodity object is requested, 
	// *	it will be provided from the freeList.  When the freeList is
	// *	depleted, it will be repopulated with new NodeCommodity objects.
	// *********************************************************************
	static NodeCommodity * freeList;

	// *********************************************************************
	// * NodeCommodity * next :
	// *	This is the next NodeCommodity oibject in the linked list of
	// *	objects.  This list is used to hold these objects within the
	// *	NodeBroker during job requirements resolution.
	// *********************************************************************
	NodeCommodity * next;
	
	// *********************************************************************
	// * ClusterNode * node :
	// *	This is a pointer to the ClusterNode object that this 
	// *	NodeCommodity represents. 
	// *********************************************************************
	ClusterNode * node;
	
	// *********************************************************************
	// * size_t nodeType :
	// *	This is the bitmask that identifies the capabilities of this
	// *	NodeCommodity object. This type information will be used to 
	// *	distribute the NodeCommodity object to requesting NodeBrokers.
	// *********************************************************************
	size_t nodeType;
	
	// *********************************************************************
	// * time_t rangeStart :
	// *	This is the time that the range associated with this 
	// *	NodeCommodity will begin.
	// *********************************************************************
	time_t rangeStart;
	
	// *********************************************************************
	// * time_t jobDuration :
	// *	This is the duration of the JOB that this NodeCommodity was 
	// *	created to accomodate.  Note that this is not the duration of
	// *	the MutualRange object associated with this NodeCommodity.
	// *********************************************************************
	time_t jobDuration;
	
	// *********************************************************************
	// * double weight:
	// *	This is the weight of the NodeCommodity object in consideration
	// *	of a jobs startTime, duration and the number of processors
	// *	required.  Periodically it may be necessary for the node's
	// *	weight to be updated in the event that it represents multiple
	// *	requirement types that require more than one processor cnt.
	// *	
	// *	It is the responsibility of the brokerage to reassign weights
	// *	when providing nodes of various types.
	// *********************************************************************
	double weight;

	// *********************************************************************
	// * NodeBroker * holder :
	// *	This is the NodeBroker object that is currently in possession
	// *	of this NodeCommodity object.
	// *********************************************************************
	NodeBroker * holder;

	// *********************************************************************
	// * NodeCommodity::NodeCommodity :
	// *	This is the protected constructor for the NodeCommodity object.
	// *	It will initialize all variables to a default value before
	// *	they are stored on the freeList for later deployment.
	// *********************************************************************
	NodeCommodity ( void );

public:
	// *********************************************************************
	// * NodeCommodity::NodeCommodity :
	// *	This is the public constructor for the NodeCommodity object.
	// *	It must be provided with a reference to a ClusterNode object
	// *	and an integer identifying it's type.  Note: that the NodeType
	// *	here is NOT an index that will be converted to a type... rather
	// *	it is a full nodeType value.
	// *
	// *	Type 0 would be 1<<0  -- or 1
	// *	Type 1 would be 1<<1  -- or 2
	// *	Type 2 would be 1<<2  -- or 4
	// *    Type 0, 1 and 2 would be 1<<0 | 1<<1 | 1<<2 -- or 7
	// *********************************************************************
	NodeCommodity (  ClusterNode & Node, size_t NodeType, time_t start, time_t duration );
	
	// *********************************************************************
	// * NodeCommodity::~NodeCommodity :
	// *	This is the local destructor for the NodeCommodity object.  
	// *	Since the object does not perform local storage of data, this
	// *	destructor is merely a placeholder.
	// *********************************************************************
	~NodeCommodity ( void );
	
	// *********************************************************************
	// * NodeCommodity::operator new :
	// *	This operator is used whenever a new NodeCommodity object is 
	// *	requested.  It will provide one of the preallocated objects
	// *	from the freeList.  If the freeList is depleted, then this
	// *	method will repopulate it.
	// *********************************************************************
	void * operator new ( size_t );
	
	// *********************************************************************
	// * NodeCommodity::operator delete :
	// *	This operator will return deleted NodeCommodity objects to the
	// *	freeList where they can be reissued whenever a new NodeCommodity
	// *	object is requested.
	// *********************************************************************
	void operator delete ( void * ptr );
	
	// *********************************************************************
	// * These are the inline accessor methods that provide access to the
	// * internal variables for this class.
	// *********************************************************************
	inline NodeCommodity * getNext   ( void )                 { return next; }
	inline void            setNext   ( NodeCommodity * Next ) { next = Next; }
	inline ClusterNode   & getNode   ( void )                 { return *node; }
	inline size_t          getType   ( void )                 { return nodeType; }
	inline double          getWeight ( void )                 { return weight; }
	inline NodeBroker    * getHolder ( void )                 { return holder; }
	inline void            setWeight ( double x )             { weight = x; }
	inline void            setHolder ( NodeBroker * Holder )  { holder = Holder; }
	inline void            addType   ( size_t type );
	inline int             hasType   ( size_t type );
		
	// *********************************************************************
	// * NodeCommodity::computeWeight :
	// *	This method will used the information provided to compute the 
	// *	weight of this node.  The nodes weight will then be used to 
	// *	establish it's position in the list of nodes.
	// *********************************************************************
	double computeWeight ( size_t procCnt );
	
	// *********************************************************************
	// * NodeCommodity::setType :
	// *	This method is called to directly change the nodeType of the
	// *	object.  This method will call a refresh on it's holder to 
	// *	generate a complete update.
	// *********************************************************************
	inline void setType ( size_t type ); 
	
	// *********************************************************************
	// * NodeCommodity::asciiDump :
	// *	This method is used to generate a diagnostic ascii report of 
	// *	the contents of the object. 
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};


// ******************************************************************************
// * class NodeBroker:
// *	The NodeBroker class is a mechanism that is used to exchange matching
// *	node commodities that are available within a mutual range.  If a 
// *	job has multiple requirements and has a list of available nodes (some
// *	of which match multiple requirements) this approach is used to best
// *	distribute the nodes to match the requirements.
// *
// *	There are two type of NodeBroker. One is the broker object, it is
// *	responsible for holding all of the matching nodes that have not been
// *	distributed to the various requirement clients.  The other is the
// *	client object, it is responsible for requesting and collecting a 
// *	specified number of nodes (from the broker) that provide the 
// *	capabilities it represents.
// *	
// *	The number of different requirement types that can be supported by a 
// *	NodeBroker is currently limited to 32 in order to be compatible with
// *	the size of an integer across all modern platforms.
// *
// *	Note also that this object is not managed by use of a freeList. A 
// *	sufficient array of these object are staticly allocated within the
// *	MutualRange object (which is never deleted).
// *****************************************************************************
class NodeBroker 
{
protected:
	// *********************************************************************
	// * NodeCommodity::addType :
	// *	This method is defined as a friend to allow it to call the
	// *	protected refreshType method to update the typeIndex of the
	// *	NodeBroker that owns it.
	// *********************************************************************
	friend void NodeCommodity::addType ( size_t );

	// *********************************************************************
	// * size_t typeMask :
	// *	This is an integer representing a bitmask that indicates
	// *	the types of capabilities that are represented by the 
	// *	NodeCommodities held by this object.
	// *********************************************************************
	size_t typeMask;
	
	// *********************************************************************
	// * size_t typeIndex :
	// *	This is an integer identifying the type of node that this 
	// *	object requires.  Notably, these values are arbitrary and are
	// *	derived from the information provided by the requestor.  At
	// *	construction time this data will be obtained from the 
	// *	RangeDefinition objects in the RangeTable.
	// *
	// *	This value will be 0 for the broker object and will only be
	// *	used by the client objects.
	// *********************************************************************
	size_t typeIndex;
	
	// *********************************************************************
	// * size_t nodeReq :
	// *	This is an integer identifying the number of nodes required
	// *	of the specified type.
	// *********************************************************************
	size_t nodeReq;

	// *********************************************************************
	// * size_t nodeCnt:
	// *	This is the number of nodes that are currently stored within
	// *	this object.
	// *********************************************************************
	size_t nodeCnt;
	
	// *********************************************************************
	// * size_t procCnt :
	// * 	This is the number of processors per node required by the 
	// *	specific type.  In the MasterBroker this value is initially 0
	// *	and is later used to track the current ordering of the 
	// *	various NodeCommodities.
	// *********************************************************************
	size_t procCnt;
	
	// *********************************************************************
	// * NodeCommodity * nodes;
	// * NodeCommodity * nodeTail;
	// *	This is a linked list of nodes that have been provided to this
	// *	NodeBroker.  The nodes variable represents the head of the list
	// *	and the nodeTail object represents the end of the list.
	// *********************************************************************
	NodeCommodity * nodes;
	NodeCommodity * nodeTail;

	// *********************************************************************
	// * NodeBroker::refreshType :
	// *	This method is called whenever the addType method is called for
	// *	one of the NodeCommodity objects that is owned by this broker.
	// *	It allows the new type to be immediately included in the list
	// *	of types that is owned by this Broker.
	// *
	// *	Since this method is being called by a trusted source -
	// *	the NodeCommodity - we're going to assume that the type
	// *	specification is valid.
	// *
	// *	NOTE : It is assumed that this method is being called with a 
	// *	       precalculated type mask and NOT a type index... 
	// *	       Consequently, the value can be directly ORed with the
	// *           existing typeMask.
	// *********************************************************************
	inline void refreshType ( size_t TypeMask )
		{
		typeMask |= TypeMask;
		}
		
public:
	// *********************************************************************
	// * NodeBroker::NodeBroker :
	// *	This is the public constructor for the object. It will 
	// *	initialize all internal variables.
	// *********************************************************************
	NodeBroker ( void );
	
	// *********************************************************************
	// * NodeBroker::~NodeBroker :
	// *	This is the destructor for the NodeBroker object.  It must
	// *	delete any NodeCommodity objects that have been allocated
	// *	to the class.
	// *********************************************************************
	~NodeBroker ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods that are defined to obtain
	// * information about the internal components of the class.
	// *********************************************************************
	inline size_t          getTypeMask  ( void )       { return typeMask; }
	inline void            setTypeIndex ( size_t idx ) { typeIndex = idx; }
	inline size_t          getTypeIndex ( void )       { return typeIndex; }
	inline void            setNodeReq   ( size_t req ) { nodeReq = req; } 
	inline size_t          getNodeReq   ( void )       { return nodeReq; }
	inline size_t          getNodeCnt   ( void )       { return nodeCnt; } 
	inline NodeCommodity * getNodes     ( void )       { return nodes; }
	inline size_t          getProcCnt   ( void )       { return procCnt; }
	inline void            setProcCnt   ( size_t x )   { procCnt = x; }

	// *********************************************************************
	// * NodeBroker::clear :
	// *	This method is provided to initialize the internals of the 
	// *	NodeBroker class during the construction of its parent class.
	// *********************************************************************
	inline void clear ( void )
		{
		typeMask  = 0;
		typeIndex = 0;
		nodeReq   = 0;
		nodeCnt   = 0;
		procCnt   = 0;
		nodes     = NULL;
		nodeTail  = NULL;
		}

	// *********************************************************************
	// * NodeBroker::satisified :
	// *	This method will check to see if the NodeBroker has obtained
	// *	a sufficient number of nodes to satisfy the job requirements.
	// *********************************************************************
	inline int satisfied ( void ) { return (nodeCnt>=nodeReq); }

	// *********************************************************************
	// * NodeBroker::hasType :
	// *	This method will check the typeMask to determine if this 
	// *	NodeBroker has a specific type of node.  It returns a boolean 
	// *	result indicating success (non-zero) or failure (zero).
	// *	
	// *	NOTE: Since the value provided in the typeIdx is an index, it
	// *	      must be bit-shifted before it can be ANDed with the
	// *	      typemask. 
	// *********************************************************************
	inline int hasType ( size_t typeIdx ) 
		{ return ((typeIdx<SchedStatic::MAXSUBSETS)?typeMask&(1<<typeIdx):0); }
	
	// *********************************************************************
	// * NodeBroker::issueNode :
	// *	This method is used to give a node to the NodeBroker object. The
	// *	method will install the NodeCommodity object to the end of the
	// *	list, will update the typeMask to represent all capabilities
	// *	that are provided by this object and will increment the 
	// *	nodeCnt variable.
	// *********************************************************************
	virtual void issueNode ( NodeCommodity & node );
	
	// *********************************************************************
	// * NodeBroker::requestNode :
	// *	This method will search through the list of nodes to find one
	// *	that represents a specific type.  If the client has this type
	// *	of node, then it will provide it to the requestor - otherwise,
	// *	a NULL is returned.
	// *********************************************************************
	virtual NodeCommodity * requestNode ( size_t typeIdx );
	
	// *********************************************************************
	// * NodeBroker::revokeNode :
	// *	This method should only be called when the type associated with
	// *	a NodeCommodity has been drastically changed.  This will walk
	// *	through the list of NodeCommodities to find the one that has
	// *	the specified index and will remove it from the list.
	// *********************************************************************
	NodeCommodity * revokeNode ( size_t nodeIdx );
	
	// *********************************************************************
	// * NodeBroker::refresh :
	// *	Because a NodeCommodity's characteristics may be modified 
	// *	externally, each of the brokers must be refreshed prior to
	// *	use to ensure that their typeMask is up to date.
	// *********************************************************************
	void refresh ( void );

	// *********************************************************************
	// * NodeBroker::asciiDump :
	// *	This method allows the caller to generate an ascii text
	// *	diagnostic dump of the contents of the NodeBroker object. 
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

// *****************************************************************************
// * class MasterBroker :
// *	This class will be used to hold nodes that may represent varying node
// *	type requirements.  If the number of processors required varies, then
// *	each NodeCommodities weight may shift --- thus altering their order.
// *	This class has the capacity to dynamically reorder it's list of nodes
// *	based on the number of processors that are being requested by the node
// *	type.
// *****************************************************************************
class MasterBroker : public NodeBroker 
{
protected:
	// *********************************************************************
	// * MasterBroker::requestNode :
	// *	This method has been made protected to prevent it from being
	// *	called.
	// *********************************************************************
	NodeCommodity * requestNode ( size_t typeIdx ) { return NULL; }
	
public:
	// *********************************************************************
	// * MasterBroker::MasterBroker :
	// *	This is the initializer for the MasterBroker object.
	// *********************************************************************
	MasterBroker ( void );

	// *********************************************************************
	// * MasterBroker::issueNode :
	// *	This method will insert the NodeCommodity into the list of 
	// *	NodeCommodities. When the node is inserted, it's weight will
	// *	be recomputed using the current procCnt value. If the weight
	// *	is valid, then it will be installed into it's proper location
	// *	in the array... otherwise, it will be appended to the end of
	// *	the list.
	// *********************************************************************
	void issueNode ( NodeCommodity & node );

	// *********************************************************************
	// * MasterBroker::requestNode :
	// *	If necessary, the nodes will be resorted based on the caller 
	// *	specified procCnt and then the 'most elligible' node of the
	// *	specified type will be returned to the caller.
	// *********************************************************************
	NodeCommodity * requestNode ( size_t typeIdx, size_t procCnt );
};


// #############################################################################
// # Out-of-class inline definitions.  These inline functions rely on features
// # of the NodeBroker class that are defined later in the source code.
// # Consequently, they must be realized at the end of the header file.
// #############################################################################

// *****************************************************************************
// * NodeCommodity::setType :
// *	This method is called to directly change the nodeType of the
// *	object.  This method will call a refresh on it's holder to 
// *	generate a complete update.
// *****************************************************************************
inline void NodeCommodity::setType ( size_t type )
	{
	nodeType=type; 
	if(holder) holder->refresh();
	}
	
// *****************************************************************************
// * NodeCommodity::addType :
// *	This method allows the caller to add a new type to the 
// *	NodeCommodity object.
// *
// *	NOTE: Since this method is being called with a type index, the type
// *	      value must be bit-shifted before it can be ORed with the nodeType.
// *****************************************************************************
inline void NodeCommodity::addType ( size_t type ) 
	{
	if(type<SchedStatic::MAXSUBSETS)
		{
		nodeType|=(1<<type); 
		if(holder) holder->refreshType(nodeType);
		}
	}

// *****************************************************************************
// * NodeCommodity::hasType :
// *	This method allows the caller to discover if the node supports
// *	a specific type.
// *
// *	NOTE: Since the type provided to the method is a type index, it must
// *	      be bit-shifted before it can be ANDed with the typeMask.
// *****************************************************************************
inline int NodeCommodity::hasType ( size_t type )
	{
	return((type<SchedStatic::MAXSUBSETS)?nodeType&(1<<type):0); 
	}

#endif /* _BROKERAGE_H_ */
