// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The file ClusterRange.cc contains the methods and attributes for the 
// 	class definitions in ClusterRange.h
// 	class.  
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <ClusterRange.h>	// Needed for class ClusterRange definitions

// *****************************************************************************
// * ClusterRange::ClusterRange :
// *	This is the constructor for the object. It will initialize all
// *	attributes to their default values.
// *****************************************************************************
ClusterRange::ClusterRange ( void )
	: listHead (NULL),
	  listTail (NULL),
	  range    (NULL),
	  nodeCnt  (0)
	{
	}
	
// *****************************************************************************
// * ClusterRange::~ClusterRange :
// *	This is the destructor for the object. It will delete all of the
// *	Range objects that have been harvested from the RangeList 
// *	objects and will then delete the RangeList objects.
// *****************************************************************************
ClusterRange::~ClusterRange ( void )
	{
	while(range!=NULL)
		{
		Range * r = range;
		range     = range->getNext();
		delete r;
		}
	while(listHead!=NULL)
		{
		RangeList * rl = listHead;
		listHead       = listHead->getNext();
		delete rl;
		}
	nodeCnt = 0;
	}

// *****************************************************************************
// * ClusterRange::incorporate :
// *	This method will add a new RangeList object to the listHead
// *	pointer. The Range objects that are stored within the RangeList
// *	object will then be extracted and placed into the ordered list
// *	of ranges that are stored locally.
// *
// *	Note:  Both the RangeList object and its contained Ranges
// *	become the property of this ClusterRange object and should
// *	not be modified or accessed externally after incorporation.
// *****************************************************************************
void ClusterRange::incorporate ( RangeList & list )
	{
	Range * rangeArray  = list.extractRanges();

	// *********************************************************************
	// * If there are no Range objects in the RangeList, then delete the
	// * list and don't include this list in the nodeCnt... Otherwise...
	// *********************************************************************
	if(rangeArray==NULL) delete &list;
	else	{
		// *************************************************************
		// * Copy the list to the tail of the listHead list.
		// *************************************************************
		if(listHead==NULL) listHead = &list;
		else               listTail->setNext(&list);
		listTail = &list;

		// *************************************************************
		// * Increment the nodeCnt variable to indicate that another 
		// * node has been added.
		// *************************************************************
		nodeCnt++;
		
		// *************************************************************
		// * Set the currEntry to point to the Range object at the head
		// * of the list. Set the prevEntry to NULL (it will be advanced
		// * later.)
		// *************************************************************
		Range * currEntry = range;
		Range * prevEntry = NULL;
		
		// *************************************************************
		// * While entries still exist in the rangeArray that was 
		// * extracted from the RangeList - integrate these objects into
		// * the local range list.
		// *************************************************************
		while(rangeArray!=NULL)
			{
			// *****************************************************
			// * Place the first/next rangeArray entry into the 
			// * newEntry pointer and then advance the rangeArray
			// * pointer to the next entry.
			// *****************************************************
			Range * newEntry = rangeArray;
			rangeArray = rangeArray->getNext();

			// *****************************************************
			// * Set the next pointer of the newEntry to NULL to 
			// * prevent subsequent entries from being added to the
			// * list at the wrong location.
			// *****************************************************
			newEntry->setNext(NULL);

			// *****************************************************
			// * Now we begin the process of walking through the
			// * list of existing Range objects to find the best
			// * place to install this object.  Note:  If the range
			// * pointer IS NULL, then the range list is currently
			// * empty and the newEntry will be the first on the 
			// * list.
			// *****************************************************
			if(range==NULL)
				{
				range     = newEntry;
				currEntry = newEntry;
				prevEntry = NULL;
				}
			// *****************************************************
			// * If currEntry has had some sort of logical disaster
			// * and ended up NULL, set currEntry to the top of the
			// * list and start over from the beginning. Remember 
			// * that it will be necessary to press the newEntry
			// * back onto the rangeArray so that it will be 
			// * processed on the next cycle.
			// *****************************************************
			else if(currEntry==NULL) 
				{
				currEntry = range;
				prevEntry = NULL;
				newEntry->setNext(rangeArray);
				rangeArray = newEntry;
				}
			// *****************************************************
			// * Here's where the action is.  We're going to walk
			// * down the list of existing entries until we get
			// * to one that is either NULL or has a start time
			// * greater than the newEntry and a shorter duration.
			// *****************************************************
			else	{
				while(currEntry && 
				      (currEntry->getStartTime()<newEntry->getStartTime()  || 
				      (currEntry->getStartTime()==newEntry->getStartTime() &&
				      (currEntry->getEndTime()>newEntry->getEndTime()))))
					{
					prevEntry = currEntry;
					currEntry = prevEntry->getNext();
					}
				// *********************************************
				// * At this point, prevEntry is either NULL 
				// * (indicating that the newEntry is earlier
				// * and/or longer than the first entry in the
				// * list) OR prevEntry points to the last
				// * entry that is earlier and/or longer than
				// * the new entry...
				// * If prevEntry is NULL - place newEntry at
				// * the top of the list.  Otherwise, install
				// * newEntry immediately after prevEntry.
				// *********************************************
				if(prevEntry==NULL) range = newEntry;
				else prevEntry->setNext(newEntry);
				
				// *********************************************
				// * As always, set the next entry after the new
				// * range to the value in the currEntry pointer
				// * to avoid breaking the linked list.
				// *********************************************
				newEntry->setNext(currEntry);
				
				// *********************************************
				// * Set currEntry to point to the newly added
				// * entry... Because the Range objects 
				// * specified in the user provided list are 
				// * already ordered, it is safe to continue
				// * from the current location.
				// *********************************************
				currEntry = newEntry;
				}
			}		
		}
	}

// *****************************************************************************
// * ClusterRange::asciiDump :
// *	This method will generate a diagnostic dump that shows the
// *	contents and ordering of this ClusterRange object.
// *****************************************************************************
void ClusterRange::asciiDump ( FILE * fp, char * prefix)
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	char * subPrefixString = new char[strlen(prefixString)+7];
	sprintf(subPrefixString, "%s      ", prefixString);
	
	fprintf(fp, "%sCluster Range :\n", prefixString);
	fprintf(fp, "%s   => Number of Nodes : %u\n", prefixString, nodeCnt);
	fprintf(fp, "%s   => Range Lists     :\n", prefixString);
	RangeList * lists = listHead;
	while(lists!=NULL)
		{
		lists->asciiDump(fp, subPrefixString);
		lists = lists->getNext();
		}
	fprintf(fp, "%s   => Ranges          :\n", prefixString);
	Range * rangeArray = range;
	while(rangeArray!=NULL)
		{
		rangeArray->asciiDump(fp, subPrefixString);
		rangeArray = rangeArray->getNext();
		}
	fflush(fp);
	delete subPrefixString;
	}	
