// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The header file ClusterRange.h contains the class definitions
// 	for the ClusterRange class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _CLUSTER_RANGE_H_
#define _CLUSTER_RANGE_H_ 1

#include <RangeList.h>		// Needed for RangeList class definition
#include <Range.h>		// Needed for Range class definitions

// *****************************************************************************
// * class ClusterRange :
// *	The ClusterRange object is a container for managing a collection of
// *	RangeList objects - each of which contains the available time Ranges
// *	from a single node.  This oibject will collect all of those Ranges
// *	together for all nodes on the cluster and will store them in an 
// *	ordered list.
// *****************************************************************************
class ClusterRange
{
protected:
	// *********************************************************************
	// * RangeList * listHead :
	// *	This is a linked list of all of the RangeList objects that have
	// *	been incorporated into this object.
	// *********************************************************************
	RangeList * listHead;
	RangeList * listTail;
	
	// *********************************************************************
	// * Range * range :
	// *	This is an ordered list of Range objects that were harvested
	// *	from the individual RangeList objects. 
	// *********************************************************************
	Range * range;
	
	// *********************************************************************
	// * size_t nodeCnt :
	// *	This is the number of RangeList objects (each representing a 
	// *	different node) that were incorporated by this ClusterRange
	// *	object.
	// *********************************************************************
	size_t nodeCnt;

public:
	// *********************************************************************
	// * ClusterRange::ClusterRange :
	// *	This is the constructor for the object. It will initialize all
	// *	attributes to their default values.
	// *********************************************************************
	ClusterRange ( void );
		
	// *********************************************************************
	// * ClusterRange::~ClusterRange :
	// *	This is the destructor for the object. It will delete all of the
	// *	Range objects that have been harvested from the RangeList 
	// *	objects and will then delete the RangeList objects.
	// *********************************************************************
	~ClusterRange ( void );

	// *********************************************************************
	// * ClusterRange::incorporate :
	// *	This method will add a new RangeList object to the listHead
	// *	pointer. The Range objects that are stored within the RangeList
	// *	object will then be extracted and placed into the ordered list
	// *	of ranges that are stored locally.
	// *
	// *	Note:  Both the RangeList object and its contained Ranges
	// *	become the property of this ClusterRange object and should
	// *	not be modified or accessed externally after incorporation.
	// *********************************************************************
	void incorporate ( RangeList & list );

	// *********************************************************************
	// * ClusterRange::getRange :
	// *	This method will return the Range linked list pointer. This 
	// *	provides the caller with full access to the Range objects, 
	// *	however, they remain the property of this class and should not
	// *	be reordered or modified outside of this class.
	// *********************************************************************
	inline Range * getRange ( void ) { return range; }

	// *********************************************************************
	// * ClusterRange::getNodeCnt :
	// *	This method returns the number of RangeList objects that have
	// *	been incorporated into this class.  Since each RangeList 
	// *	represents a single node, this value (in effect) identifies the
	// *	number of nodes that have time ranges specified.
	// *********************************************************************
	inline size_t getNodeCnt ( void ) { return nodeCnt; }

	// *********************************************************************
	// * ClusterRange::asciiDump :
	// *	This method will generate a diagnostic dump that shows the
	// *	contents and ordering of this ClusterRange object.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL);
};

#endif /* _CLUSTER_RANGE_H_ */
