// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	FreeTimeSlot.cc realizes the methods defined within the FreeTimeSlot
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <FreeTimeSlot.h>
#include <SchedStatic.h>
#include <string.h>

// *****************************************************************************
// * FreeTimeSlot::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the FreeTimeSlot class.
// *****************************************************************************
FreeTimeSlot * FreeTimeSlot::freeList;

// *****************************************************************************
// * operator FreeTimeSlot::new :
// *	This operator is called whenever a new FreeTimeSlot object is 
// *	requested. The operator will first attempt to extract a FreeTimeSlot
// *	object from those preallocated on the freeList.  If no FreeTimeSlot
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * FreeTimeSlot::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new FreeTimeSlot[32];
		for(i=0; i<31; i++) freeList[i].next = &freeList[i+1];
		freeList[i].next = NULL;
		}

	FreeTimeSlot * retval = freeList;
	freeList     = retval->next;
	retval->next = NULL;
	
	return (void *)retval;
	}

	
// *****************************************************************************
// * operator FreeTimeSlot::delete :
// *	This is the delete operator for the FreeTimeSlot class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void FreeTimeSlot::operator delete ( void * ptr )
	{
	((FreeTimeSlot *)ptr)->next = freeList;
	freeList = (FreeTimeSlot *)ptr;
	}
	
	
// *****************************************************************************
// * FreeTimeSlot::FreeTimeSlot : 
// *	This is a private constructor for the FreeTimeSlot object.  It will
// *	be called each time the FreeTimeSlot::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
FreeTimeSlot::FreeTimeSlot ( void )
	: startTime(0), duration(0), next (NULL)
	{
	}
	
// *****************************************************************************
// * FreeTimeSlot::FreeTimeSlot :
// *	This is the public constructor for the FreeTimeSlot object. It will
// *	be called when the user needs to allocate a new FreeTimeSlot object.
// *****************************************************************************
FreeTimeSlot::FreeTimeSlot ( time_t StartTime, time_t Duration )
	: startTime ( StartTime ),
	  duration  ( Duration ),
	  next      ( NULL)
	{
	}
	
// *****************************************************************************
// * FreeTimeSlot::~FreeTimeSlot :
// *	This is the destructor for the FreeTimeSlot object.  It will deallocate
// *	the memory that was allocated to the name string and will clear all 
// *	variables before the object is returned to the freeList.
// *****************************************************************************
FreeTimeSlot::~FreeTimeSlot ( void )
	{
	startTime = 0;
	duration  = 0;
	next      = NULL;
	}
	
// *****************************************************************************
// * FreeTimeSlot::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the FreeTimeSlot object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void FreeTimeSlot::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
		
	fprintf(fp, "%sFree time starts at %s and runs %i:%02i:%02i hours\n", 
		prefixString, SchedStatic::getTimeString(startTime), 
		duration/3600, (duration%3600)/60, duration%60);
	}
	
