// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distributioon.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	JobAssignment.cc realizes the methods defined within the JobAssignment
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <JobAssignment.h>
#include <SchedStatic.h>
#include <stdlib.h>
#include <string.h>

// *****************************************************************************
// * JobAssignment::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the JobAssignment class.
// *****************************************************************************
JobAssignment * JobAssignment::freeList;

// *****************************************************************************
// * operator JobAssignment::new :
// *	This operator is called whenever a new JobAssignment object is 
// *	requested. The operator will first attempt to extract a JobAssignment
// *	object from those preallocated on the freeList.  If no JobAssignment
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * JobAssignment::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new JobAssignment[32];
		for(i=0; i<31; i++) freeList[i].setNext(&freeList[i+1]);
		freeList[i].setNext(NULL);
		}
	JobAssignment * retval = freeList;
	freeList = retval->getNext();
	retval->setNext(NULL);
	
	return (void *)retval;
	}

	
// *****************************************************************************
// * operator JobAssignment::delete :
// *	This is the delete operator for the JobAssignment class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void JobAssignment::operator delete ( void * ptr )
	{
	((JobAssignment *)ptr)->next = freeList;
	freeList = (JobAssignment *)ptr;
	}

	
// *****************************************************************************
// * JobAssignment::JobAssignment:
// *	This is a private constructor for the JobAssignment object.  It will
// *	be called each time the jobAssignment::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
JobAssignment::JobAssignment ( void )
	: name(NULL), startTime(0), duration(0), next (NULL)
	{
	}
	
// *****************************************************************************
// * JobAssignment::JobAssignment :
// *	This is the public constructor for the JobAssignment object. It will
// *	be called when the user needs to allocate a new JobAssignment object.
// *****************************************************************************
JobAssignment::JobAssignment ( char * Name, time_t StartTime, time_t Duration )
	: name      (Name?strdup(Name):NULL),
	  startTime ( StartTime ),
	  duration  ( Duration ),
	  next      ( NULL)
	{
	}
	
// *****************************************************************************
// * JobAssignment::~JobAssignment :
// *	This is the destructor for the JobAssignment object.  It will deallocate
// *	the memory that was allocated to the name string and will clear all 
// *	variables before the object is returned to the freeList.
// *****************************************************************************
JobAssignment::~JobAssignment ( void )
	{
	if(name) free(name);
	name      = NULL;
	startTime = 0;
	duration  = 0;
	next      = NULL;
	}
	
// *****************************************************************************
// * JobAssignment::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the JobAssignment object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void JobAssignment::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
	fprintf(fp, "%sJob %s starts at %s and runs %i:%02i:%02i hours\n", 
		prefixString, name?name:"UNKNOWN", 
		SchedStatic::getTimeString(startTime),
		duration/3600, (duration%3600)/60, duration%60);
	}
	
