// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	JobAssignment.h contains the class definitions for the JobAssignment
// 	class used to maintain a list of running and scheduled jobs within each
//	processor.
// 
//  Author:  Walt Akers
// ------------------------------------------------------------------------

#ifndef _JOB_ASSIGNMENT_H_
#define _JOB_ASSIGNMENT_H_ 1

#include <stdio.h>
#include <time.h>

// *****************************************************************************
// * class JobAssignment :
// *	The JobAssignment class maintains a list of all jobs that are currently 
// *	running or scheduled to run on a specific processor of a node.
// *
// *	The JobAssignment object maintains the name, starttime, and duration of 
// *	each job that has been assigned to the processor.
// *
// *	In order to reduce memory allocations, the JobAssignment object 
// *	maintains a static freelist of these objects.  Whenever a new 
// *	JobAssignment object is required, a preallocated one will be supplied 
// *	from the freelist. If the freelist is depleted, an array of 32 
// *	JobAssignment objects will be allocated to repopulate it.  As 
// *	JobAssignment objects are deleted by their users, they are returned to 
// *	the freelist to be reused at a later date.
// *****************************************************************************
class JobAssignment
{
private:
	// *********************************************************************
	// * The freeList pointer is used to maintain a list of preallocated
	// * JobAssignment objects that may be provided to the requestor.  This
	// * approach reduces the number of dynamic allocations and improves
	// * overall performance.
	// *********************************************************************
	static JobAssignment * freeList;

	JobAssignment * next;
	char          * name;
	time_t          startTime;
	time_t          duration;

	// *********************************************************************
	// * This is a private constructor that is used to populate the
	// * freeList.
	// *********************************************************************
	JobAssignment ( void );

public:
	// *********************************************************************
	// * New and delete operators used to allocate the JobAssignment object
	// * from the freeList.
	// *********************************************************************
	void * operator new    ( size_t Size );
	void   operator delete ( void * ptr );

	// *********************************************************************
	// * Class constructor/destructor method definitions.
	// *********************************************************************
	JobAssignment  ( char * Name, time_t StartTime, time_t Duration);
	~JobAssignment ( void );

	// *********************************************************************
	// * Inline data accessor methods.
	// *********************************************************************
	const           char * getName ( void ) { return (const char *)name; }
	time_t          getStartTime   ( void ) { return startTime; }
	time_t          getEndTime     ( void ) { return startTime+duration; }
	time_t          getDuration    ( void ) { return duration; }
	JobAssignment * getNext        ( void ) { return next; }
	void            setNext        ( JobAssignment * Next ) { next = Next; }

	// *********************************************************************
	// * Data reporting method used for generating diagnostic dumps.
	// ********************************************************************* 
	void asciiDump ( FILE *fp=stdout, char *prefix=NULL );
};

#endif /* _JOB_ASSIGNMENT_H_ */
