// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distributioon.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	MutualRange.cc realizes the methods defined within the MutualRange
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <MutualRange.h>	// Needed for MutualRange class definition
#include <RangeTable.h>		// Needed for RangeTable class definition
#include <Range.h>		// Needed for Range class definition
#include <JobCandidate.h>	// Needed for JobCandidate class definition
#include <RangeDefinition.h>	// Needed for RangeDefinition class definition
#include <ClusterNode.h>	// Needed for ClusterNode definition
#include <JobDeployment.h>	// Needed for JobDeployment definition
#include <PBS_Log.h>		// Needed for LogError definition

// *****************************************************************************
// * MutualRange::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the MutualRange class.
// *****************************************************************************
MutualRange * MutualRange::freeList;

// *****************************************************************************
// * operator MutualRange::new :
// *	This operator is called whenever a new MutualRange object is 
// *	requested. The operator will first attempt to extract a MutualRange
// *	object from those preallocated on the freeList.  If no MutualRange
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * MutualRange::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new MutualRange[32];
		for(i=0; i<31; i++) freeList[i].setNext(&freeList[i+1]);
		freeList[i].setNext(NULL);
		}
	MutualRange * retval = freeList;
	freeList = retval->getNext();
	retval->setNext(NULL);
	
	return (void *)retval;
	}
	
// *****************************************************************************
// * operator MutualRange::delete :
// *	This is the delete operator for the MutualRange class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void MutualRange::operator delete ( void * ptr )
	{
	((MutualRange *)ptr)->setNext(freeList);
	freeList = (MutualRange *)ptr;
	}
	
	
// *****************************************************************************
// * MutualRange::MutualRange:
// *	This is a private constructor for the MutualRange object.  It will
// *	be called each time the MutualRange::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
MutualRange::MutualRange ( void )
	: rangeStart   (0), 
	  rangeDuration(0), 
	  owner        (NULL),
	  next         (NULL)
	{
	}
	
// *****************************************************************************
// * MutualRange::addNode :
// *	This method is called by the MutualRange object when it needs to 
// *	duplicate its contents in another MutualRange. If a NodeCommodity
// *	object that relates to this object exists, then it will be updated.
// *****************************************************************************
void MutualRange::addNode ( NodeCommodity & node )
	{
	size_t index = node.getNode().getIndex();
	
	// *********************************************************************
	// * If an associated NodeCommodity does not exist... create it.
	// *********************************************************************
	if(nodes[index]==NULL)
		{
		nodes[index] = new NodeCommodity
			(node.getNode(), node.getType(), rangeStart, 
			 owner->getJob()->getDuration());

		broker.issueNode(*nodes[index]);
		}
	// *********************************************************************
	// * If the NodeCommodity exists and it is different...
	// *********************************************************************
	else if(nodes[index]->getType()!=node.getType())
		{
		// *************************************************************
		// * Find out who owns it.
		// *************************************************************
		NodeBroker * holder = nodes[index]->getHolder();
		
		// *************************************************************
		// * If someone has possession of it, then take it away.
		// *************************************************************
		if(holder) holder->revokeNode(index);

		// *************************************************************
		// * Call setType to modify the type to what is specified in the
		// * NodeCommodity parameter.
		// *************************************************************
		nodes[index]->setType(node.getType());

		// *************************************************************
		// * Return it to the broker for later distribution.
		// *************************************************************
		broker.issueNode(*nodes[index]);
		}
	}
	
// *****************************************************************************
// * MutualRange::addNode :
// *	Once a range has been verified using the add method, this
// *	method will be called (internally) to add the node and its
// *	type specification to the MutualRange.
// *****************************************************************************
void MutualRange::addNode ( ClusterNode & node, size_t type )
	{
	size_t index = node.getIndex();

	// *********************************************************************
	// * Note:  If the type value is out of range... Then set it to 0.
	// * This isn't the best idea in the world - but, hey, its a protected
	// * function... If it gets a squirrelly value - then its my own
	// * doing.
	// *********************************************************************
	if(type>=SchedStatic::MAXSUBSETS || type>=owner->getTypeCnt()) type=0;
	
	// **********************************************************************
	// * If an associated NodeCommodity for this ClusterNode has not been
	// * created, then allocate one, install it in the array and issue it
	// * to the broker for later distribution.  Note that the type must
	// * be converted to its final format before being passed to the 
	// * constructor.
	// **********************************************************************
	if(nodes[index]==NULL) 
		{
		nodes[index] = new NodeCommodity
			(node, (1<<type), rangeStart, 
			 owner->getJob()->getDuration());

		broker.issueNode(*nodes[index]);
		}
	// *********************************************************************
	// * If the node exists, then add the new type to it.  This will cause
	// * it to automatically update it's associated NodeBroker.
	// *********************************************************************
	else nodes[index]->addType(type);
	}
	
// *****************************************************************************
// * MutualRange::split :
// *	If a range is added that is longer or shorter than the currently
// * 	defined MutualRange parameters, it will be necessary to split
// *	the MutualRange into two parts.  This method will cause the
// *	MutualRange to be split at the specified time.
// *****************************************************************************
void MutualRange::split ( time_t splitTime )
	{
	// *********************************************************************
	// * Only generate a split if the specified time is within range.
	// *********************************************************************
	if(splitTime>rangeStart &&
	   splitTime<rangeStart+rangeDuration)
		{
		// *************************************************************
		// * Preserve the rangeDuration for later use.
		// *************************************************************
		long          oldDuration = rangeDuration;
		MutualRange * newRange    = NULL;

		// *************************************************************
		// * Modify the rangeDuration to be the time from the rangeStart
		// * to the beginning of the split.
		// *************************************************************
		rangeDuration = splitTime-rangeStart;

		// *************************************************************
		// * Create a new MutualRange object that represents the period
		// * from the split time to the remainder of the original 
		// * rangeDuration.
		// *************************************************************
		if((newRange=insert(splitTime, oldDuration-rangeDuration))!=NULL)
			{
			// *****************************************************
			// * Add a NodeCommodity object to the new MutualRange, 
			// * for each of the NodeCommodity objects that exists 
			// * in the current MutualRange.
			// *****************************************************
			for(int idx=0; idx<SchedStatic::MAXNODES; idx++)
				{
				if(nodes[idx]!=NULL)
					{
					// *************************************
					// * This is a specialized addNode 
					// * method that makes an exact 
					// * duplicate of an existing 
					// * NodeCommodity.
					// *************************************
					newRange->addNode(*nodes[idx]);
					}
				}
			}
		}
	}

// *****************************************************************************
// * MutualRange::insert :
// *	This method is called to insert a new MutualRange object AFTER this
// *	MutualRange object.  Note, I'm using an integer for duration in order
// *	to detect a negative duration on systems where the time_t value is
// *	represented by an unsigned integer.
// *****************************************************************************
MutualRange * MutualRange::insert ( time_t start, long duration )
	{
	MutualRange * newRange = NULL;	

	// *********************************************************************
	// * Check to insure that the time span represented by the new 
	// * MutualRange does not overlap with the next MutualRange.
	// *********************************************************************
	if(next)
		{
		// *************************************************************
		// * If the time periods overlap, decrease the duration of the
		// * new MutualRange to end when the other begins.
		// *************************************************************
		if(next->getStartTime() < (start+duration))
			duration = start - next->getStartTime();
		}
	// *********************************************************************
	// * If the duration of the new MutualRange is greater than 0, then
	// * insert it after this MutualRange object.
	// *********************************************************************
	if(duration>0)
		{
		newRange = new MutualRange(*owner, start, duration);
		newRange->setNext(next);
		next = newRange;
		}
	// *********************************************************************
	// * Return the newly created MutualRange object - or NULL if no
	// * object was created.
	// *********************************************************************
	return newRange;
	}
	
// *****************************************************************************
// * MutualRange::brokerExchange :
// *	This method is called to attempt to locate available 
// *	NodeCommodity objects with matching types.  If the number of
// *	commodity objects in question are located, they will 
// *	automatically be passed to the client at the specified index 
// *	and a non-zero value will be returned.  If an available, 
// *	matching number of commodity objects cannot be located - then 0 
// *	will be returned indicating that the job cannot be satisfied.
// *****************************************************************************
int MutualRange::brokerExchange ( size_t type, size_t cnt )
	{
	int             failure  = 0;
	NodeCommodity * nc1      = NULL;
	NodeCommodity * nc2      = NULL;
	NodeCommodity * nc3      = NULL;
	size_t          lastType = owner->getTypeCnt();
	size_t          idx;
	 	
	// *********************************************************************
	// * Protect against overflow by ensuring that the lastType variable
	// * cannot be greater than the maximum number of sub sets.
	// *********************************************************************
	if(lastType>SchedStatic::MAXSUBSETS) lastType = SchedStatic::MAXSUBSETS;
	
	// *********************************************************************
	// * First consideration.  If the broker does not have sufficient 
	// * commodities on board to satisfy the request (even if they are
	// * of the wrong type), then this job cannot be satisfied.  Check
	// * to see if the broker is sufficiently loaded.
	// *********************************************************************
	if(broker.getNodeCnt()<cnt) failure = 1;
	else	{
		// *************************************************************
		// * Extract all matching nodes possible from the broker. 
		// *************************************************************
		while (cnt>0 && broker.hasType(type))
			{
			// *****************************************************
			// * Extract the node from the broker... If successful
			// * then give the node to the client.  Otherwise, 
			// * refresh the broker in an attempt to clear the error
			// * that made it think that it had a type that it 
			// * didn't.
			// *****************************************************
			if((nc1=broker.requestNode(type, client[type].getProcCnt()))!=NULL)
				{
				client[type].issueNode(*nc1);
				cnt--;
				}
			else broker.refresh();
			}
	
		// *************************************************************
		// * If the broker has been exhausted and we still do not have 
		// * enough nodes of the specified type.  It's time to start 
		// * trading commodities between clients.
		// *
		// * Ask each of the clients represented by the broker's types 
		// * if they have a node of the desired type that they are 
		// * willing to trade.
		// *
		// * NOTE: Since the idx represents a type index, it must be
		// *       bitshifted before it can be ANDed with the typeMask.
		// *************************************************************
		for(idx=0; cnt>0 && idx<lastType; idx++)
			{
			while(cnt && ((broker.getTypeMask())&(1<<idx)) && 
			      client[idx].hasType(type))
				{
				// *********************************************
				// * Obtain a pointer to the 'idx' type from the
				// * broker.
				// *********************************************
				nc1 = broker.requestNode(idx, client[idx].getProcCnt());
				
				// *********************************************
				// * Obtain a pointer to the desired type from
				// * client[idx].
				// *********************************************
				nc2 = client[idx].requestNode(type);

				// *********************************************
				// * If both of the objects retrieved represent
				// * valid pointers, then feed them to the 
				// * appropriate client...
				// *********************************************
				if(nc1 && nc2)
					{
					client[idx].issueNode(*nc1);
					client[type].issueNode(*nc2);
					cnt--;		       
					}
				// *********************************************
				// * If one or more of the objects retrieved 
				// * was invalid, then place them back into 
				// * their respective sources and call refresh
				// * in an attempt to clear the problem.
				// *********************************************
				else	{
					if(nc1) broker.issueNode(*nc1);
					if(nc2) client[idx].issueNode(*nc2);
					broker.refresh();
					client[idx].refresh();
					}
				}
			}
	
		// *************************************************************
		// * If we have walked through all of the clients that are 
		// * represented as available in the broker.  It's time to start 
		// * examining the possibility of a two way exchange to satisfy 
		// * the job.
		// *************************************************************
		if(cnt > 0)
			{
			size_t otherTypes = 0;
		
			// *****************************************************
			// * Create a bitmask representing all types that should 
			// * be represented in this MutualRange.
			// *
			// * NOTE: Since the value of idx represents a type 
			// *       index, it must be bitshifted before it can
			// *       be ORed with the otherTypes.
			// *****************************************************
			for(idx=0; idx<lastType; idx++) otherTypes|=(1<<idx);

			// *****************************************************
			// * From this bitmask - remove all of the types that 
			// * are currently represented in the broker (because we
			// * know that they don't have what we want) and remove
			// * the type that we are looking to populate (because
			// * it can't afford to loose anything).
			// *
			// * NOTE: Since the value of type represents a type 
			// *       index, it must be bitshifted before it can
			// *       be ORed with the type mask.
			// *****************************************************
			otherTypes^=(broker.getTypeMask()|(1<<type));
	
			// *****************************************************
			// * Now... Walk through each of the OTHER types and see
			// * if they have one of the type needed... If they do, 
			// * see if one of the types represented in the broker
			// * has something to trade with them.
			// *****************************************************
			for(idx=0; cnt>0 && idx<lastType; idx++)
				{
				// *********************************************
				// * If idx points to one of the other types.
				// *
				// * NOTE: Since idx is a type index, it must
				// *       be bit-shifted before it can be
				// *       ANDed with the otherTypes.
				// *********************************************
				if(otherTypes&(1<<idx))
					{
					// *************************************
					// * Walk through each node and query
					// * those that are represented within
					// * the broker to see if they have a
					// * node that they can trade with an
					// * interim client to obtain the node
					// * type needed.
					// *************************************
					for(int j=0; cnt>0 && j<lastType; j++)
						{
						// *****************************
						// * NOTE: Since j represents a
						// * type index, it must be 
						// * bit-shifted before it can 
						// * be compared with the 
						// * typeMask.
						// *****************************
						while(cnt>0 && 
						      ((broker.getTypeMask())&(1<<j)) &&
						      client[j].hasType(idx) && 
						      client[idx].hasType(type))
							{
							// *********************
							// * Extract each of the
							// * desired nodes to
							// * a commodity pointer.
							// *********************
							nc1 = broker.requestNode(j, client[j].getProcCnt());
							nc2 = client[j].requestNode(idx);
							nc3 = client[idx].requestNode(type);
							
							// *********************
							// * If the items are all 
							// * valid, then distribute
							// * them to the right
							// * client.
							// *********************
							if( nc1 && nc2 && nc3 )
								{
								client[j].issueNode(*nc1);
								client[idx].issueNode(*nc2);
								client[type].issueNode(*nc3);
								cnt--;
								}
							// *********************
							// * Otherwise, put them
							// * back and refresh the
							// * objects to clear the
							// * problem.
							// *********************
							else	{
								if(nc1) broker.issueNode(*nc1);
								if(nc2) client[j].issueNode(*nc2);
								if(nc3) client[idx].issueNode(*nc3);
								broker.refresh();
								client[j].refresh();
								client[idx].refresh();
								}
							}
						}
					}
				}
			}
		}
	// *********************************************************************
	// * If cnt has been decremented to zero, then return TRUE - indicating
	// * that all needed nodes have been retrieved.  If cnt is still greater
	// * than zero, then return 0 indicating that not all needed nodes were
	// * obtained.
	// *********************************************************************
	return !(cnt);
	}

// *****************************************************************************
// * MutualRange::MutualRange :
// *	This is the public constructor for the MutualRange object. It will
// *	be called when the user needs to allocate a new MutualRange object.
// *****************************************************************************
MutualRange::MutualRange ( RangeTable & Owner, time_t StartTime, time_t Duration)
	: next         (NULL),
	  owner        (&Owner),
	  rangeStart   (StartTime), 
	  rangeDuration(Duration)
	{
	int idx;

	broker.clear();
	for(idx=0; idx<SchedStatic::MAXSUBSETS; idx++) 
		{
		RangeDefinition * ownerType = owner->getType(idx);
		client[idx].clear();
		client[idx].setTypeIndex(idx);
		client[idx].setNodeReq  (ownerType?ownerType->getNodesReq():0);
		client[idx].setProcCnt  (ownerType?ownerType->getProcsPerNode():0);
		}
	for(idx=0; idx<SchedStatic::MAXNODES; idx++) nodes[idx] = NULL;
	}
		
// *****************************************************************************
// * MutualRange::~MutualRange :
// *	This is the destructor for the MutualRange object. It will delete each 
// *	of the NodeCommodity objects that was allocated during the objects use.
// *****************************************************************************
MutualRange::~MutualRange ( void )
	{
	int idx;
	for(int idx=0; idx<SchedStatic::MAXNODES; idx++)
		{
		if(nodes[idx]!=NULL) delete nodes[idx];
		}
	}
	
// *****************************************************************************
// * MutualRange::add :
// *	This method will walk through the list of MutualRange objects and will 
// *	find the correct location to add a new Range.  The method will return a
// *	pointer to the first MutualRange object where the Range object was 
// *	utilized. If the Range object could not be used, then a pointer to the
// *	current MutualRange object will be returned.
// *
// *	Note:  If the type specified for the new range is greater than 0 - then
// *	       a new MutualRange will NOT be created to accomodate it.  This is
// *           because there is no way that the new MutualRange would ever be
// *	       able to satisfy the requirement for the nodes of earlier types.
// *****************************************************************************
MutualRange * MutualRange::add ( Range & range, size_t type )
	{	
	// *********************************************************************
	// * The type value will be used as an index.  If, for some reason the
	// * type is greater than the number of types specified in the 
	// * RangeTable or the type is greater than SchedStatic::MAXSUBSETS then
	// * the range cannot be used.
	// *********************************************************************
	if(type>=owner->getTypeCnt() || type>=SchedStatic::MAXSUBSETS)
		{
		return this;
		}
		
	// *********************************************************************
	// * This is a pointer to the first matching MutualRange object.
	// *********************************************************************
	MutualRange * firstMatch = NULL;
	
	// *********************************************************************
	// * Extract the start and duration times of this range. Note, we are 
	// * using a signed long integer for the duration in order to protect
	// * against substraction errors on systems where time_t is represented
	// * by a unsigned int.
	// *********************************************************************
	time_t start    = range.getStartTime();
	long   duration = range.getDuration();

	// *********************************************************************
	// * If Range's start is greater than the start time of this MutualRange
	// * object - then advance it to the start time of the MutualRange
	// * and decrease the duration accordingly.
	// *********************************************************************
	if(start < rangeStart)
		{
		duration -= (rangeStart-start);
		start     = rangeStart;
		}

	// *********************************************************************
	// * If after applying the start time correction, the duration of the 
	// * new range is greater than 0 and the start time of the range is less
	// * than the end time of this MutualRange, then the object can be 
	// * assimilated.
	// *********************************************************************
	if(duration>0 && start<rangeStart+rangeDuration)
		{
		// *************************************************************
		// * IF the new range completely overlaps this MutualRange,
		// * simply add it.
		// *************************************************************
		if(start==rangeStart && duration>=rangeDuration)
			{
			addNode(range.getNode(), type);
			firstMatch = this;
			}

		// *************************************************************
		// * IF the startTime of the new range is greater than the
		// * startTime of this range - Break the MutualRange at the 
		// * range's start time.
		// *
		// * Note the job will be added to the newly created next
		// * MutualRange later.
		// *************************************************************
		else if(start>rangeStart)
			{
			split(start);
			}
			
		// *************************************************************
		// * IF the new range's duration DOES NOT last to the end of 
		// * this MutualRange's duration - Break this MutualRange at the 
		// * range's duration and add the range to the current 
		// * MutualRange.
		// ************************************************************* 
		else if (duration<rangeDuration)
			{
			split(rangeStart+duration);
			addNode(range.getNode(), type);
			firstMatch = this;
			}
		}

	// *********************************************************************
	// * Now... If the duration of the Range object extends beyond the
	// * current MutualRange (this situation may have been generated by
	// * the above operations)... Add it to the subsequent MutualRanges.
	// *********************************************************************
	if(start+duration > rangeStart+rangeDuration)
		{
		// *************************************************************
		// * These two operations which result in the creation of a new
		// * MutualRange object will only be executed IF the type is
		// * 0.  If the type is greater than 0, then any newly added 
		// * object is guaranteed to never get the type 0 components
		// * that it needs to be satisfied.
		// *************************************************************
		if (type==0)
			{
			// *****************************************************
			// * If there is no next object and the type of this 
			// * range is zero... then add a new MutualRange object.
			// *****************************************************
			if(!next)
				{
				time_t nextStart = rangeStart+rangeDuration;
				time_t nextDuration = (start+duration)-nextStart;
				insert(nextStart, nextDuration);
				}
			// *****************************************************
			// * Otherwise, detect if there is a gap between the 
			// * next time range and this one.  If a gap exists, a 
			// * new MutualRange object will need to be created to 
			// * fill it.
			// *****************************************************
			else if(next->getStartTime() > rangeStart+rangeDuration)
				{
				time_t nextStart = rangeStart+rangeDuration;
				time_t nextDuration = next->getStartTime()-nextStart;
				insert(nextStart, nextDuration);
				}
			}
		// *************************************************************
		// * If a next object exists and overlaps with the Range object
		// * then call add to insert the node information.
		// *************************************************************
		if(next && next->getStartTime()<start+duration)
			{
			if(firstMatch) next->add (range, type);
			else firstMatch = next->add (range, type);
			}
		}
	
	// *********************************************************************
	// * Return the address of the first MutualRange where the Range
	// * object was added, or the address of this MutualRange if the
	// * Range object was without a match.
	// *********************************************************************
	return (firstMatch?firstMatch:this);
	}
	
// *****************************************************************************
// * MutualRange::satified :
// *	This is an interim test to ensure that the MutualRange has satisfied the
// *	job requirements upto and including the node type specified by typeIdx. 
// *	If the job has not satisfied its early requirements now, then there is 
// *	no way that it will satisfy them later.
// *****************************************************************************
int MutualRange::satisfied ( size_t typeIdx )
	{
	int isSatisfied = 1;

	// *********************************************************************
	// * If the typeIdx is outside the scope of types required by the
	// * job, then reduce it to the maximum value.
	// *********************************************************************
	if(typeIdx>=SchedStatic::MAXSUBSETS) typeIdx = SchedStatic::MAXSUBSETS-1;
	if(typeIdx>=owner->getTypeCnt()) typeIdx = owner->getTypeCnt()-1;

	// *********************************************************************
	// * Walk through each typeIdx and determine if it has been fully 
	// * satisfied.  Terminate the loop when you get to the last specified
	// * typeIdx -OR- when one of the clients IS NOT satisfied.
	// *********************************************************************
	for(size_t idx=0; idx<=typeIdx && isSatisfied; idx++)
		{
		// *************************************************************
		// * Here we check to see if we have a sufficient number of
		// * nodes to satisfy the requirement... If not, we'll need
		// * to enter into a brokerage agreement.
		// *************************************************************
		if(client[idx].getNodeReq()>client[idx].getNodeCnt())
			{
			isSatisfied = brokerExchange
				(idx, 
				client[idx].getNodeReq()-client[idx].getNodeCnt());
			}
		}
	
	return isSatisfied;
	}
		
// *****************************************************************************
// * MutualRange::claim :
// *	This method is called to transform the MutualRange object into a usable 
// *	JobDeployment object.  This method will generate the empty JobDeployment
// *	object and will then populate it by claiming the nodes that are 
// *	currently collected in the subset specific NodeBroker clients.
// *****************************************************************************
JobDeployment * MutualRange::claim ( void )
	{
	JobDeployment * result  = NULL;
	size_t          reqProc = 0;
	size_t          idx;
	
	// *********************************************************************
	// * First, count the total number of processors required by the job.
	// * This value SHOULD be specified when generating the JobDeployment 
	// * object.  While walking the list, ensure that this MutualRange has, 
	// * indeed, been satisfied.
	// *********************************************************************
	for(idx=0; idx<owner->getTypeCnt(); idx++)
		{
		RangeDefinition * def = owner->getType(idx);
		reqProc += def?(def->getProcsPerNode()*def->getNodesReq()):0;
		
		if(client[idx].getNodeCnt()<client[idx].getNodeReq() || !def)
			return NULL;
		}
		
	// *********************************************************************
	// * At this point we have derived the maximum number of procs, we
	// * know that all of the required type definitions exist and we know
	// * that we have located a sufficient number of nodes to perform
	// * the job. 
	// *
	// * It is time to generate a JobDeployment object that brings all of 
	// * that together. We'll use the start time assiociated with this 
	// * MutualRange object and the duration associated with the Job
	// * itself.
	// *********************************************************************
	time_t jobDuration   = owner->getJob()->getDuration();
	char * jobName       = (char *)owner->getJob()->getName();
	result = new JobDeployment (rangeStart, 
				    jobDuration,
				    jobName,
				    reqProc);
				    
	// *********************************************************************
	// * Now we have a JobDeployment object... Lets walk through the various
	// * NodeBroker clients and collect processors that can be used for this
	// * job. 
	// *
	// * Note:  In theory if the JobDeployment FAILS to obtain any of the 
	// * nodes that have been promised to it, then it indicates a failure
	// * in logic (because the node should have been available).  This will
	// * muddy the stream for all future jobs...  During further 
	// * development we'll need a way to unclaim a node in the event that
	// * node acquisition failed --- or we'll need to wait to the next 
	// * scheduling cycle to continue processing.
	// *********************************************************************
	for(idx=0; idx<owner->getTypeCnt(); idx++)
		{
		NodeCommodity * nlist = client[idx].getNodes();
		reqProc = owner->getType(idx)->getProcsPerNode();
		
		// *************************************************************
		// * Walk through each NodeCommodity in the list and claim the
		// * processors that are needed for this job.
		// *************************************************************
		while(nlist!=NULL)
			{
			size_t  pIdx = 0;
			char ** procs = nlist->getNode().claim
				( jobName, reqProc, rangeStart, jobDuration );

			// *****************************************************
			// * Add each processor that was collected to the 
			// * JobDeployment object.
			// *****************************************************
			if(procs) for(pIdx=0; procs[pIdx] && pIdx<reqProc; pIdx++) 
				{
				result->addProc(procs[pIdx]);
				}
			// *****************************************************
			// * It is the responsibility of the caller to delete
			// * the array returned by ClusterNode::claim - but, not
			// * the strings that are pointed to.
			// *****************************************************
			delete [] procs;

			// *****************************************************
			// * Here is where some more serious error handling 
			// * should be instituted... If the pIdx is not equal to
			// * the required number of processors, then something
			// * is wrong and will keep getting wronger as time
			// * goes by.
			// *
			// * For now, delete the JobDeployment object and return
			// * NULL... In future development we need to institute
			// * a ClusterNode::unclaim method that can release a
			// * process's association with a job.
			// *****************************************************
			if(pIdx!=reqProc)
				{
				delete result;

				LogError("MutualRange", 
					 "ERROR: Job %s didn't get all promised processors", 
					 jobName);

				return NULL;
				}
			
			// *****************************************************
			// * Advance the NodeCommodity list to the next entry.
			// *****************************************************
			nlist = nlist->getNext();
			}
		}
	
	// *********************************************************************
	// * Return the resultant JobDeployment object.
	// *********************************************************************
	return result;
	}
	
	
// *****************************************************************************
// * MutualRange::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the MutualRange object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void MutualRange::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	char * subPrefixString = new char[strlen(prefixString)+7];
	
	sprintf(subPrefixString, "%s      ", prefixString);
	
	fprintf(fp, "%sMutualRange for Job %s\n", 
		prefixString, owner->getJob()->getName());
	fprintf(fp, "%s   => Time Range Start     : %s\n", 
		prefixString, SchedStatic::getTimeString(rangeStart));
	fprintf(fp, "%s   => Time Range End       : %s\n",
		prefixString, SchedStatic::getTimeString(rangeStart+rangeDuration));
	fprintf(fp, "%s   => Number of Sub-Types  : %i\n", 
		prefixString, owner->getTypeCnt());
	for(int idx=0; idx<owner->getTypeCnt(); idx++)
		{
		RangeDefinition * typeDef = owner->getType(idx);
		
		fprintf(fp, "%s   => Type %i\n", prefixString, idx);
		fprintf(fp, "%sRequirements :\n", subPrefixString);
		if(typeDef) typeDef->asciiDump(fp, subPrefixString);
		fprintf(fp, "%s\n%sNodes Assigned :\n", subPrefixString, subPrefixString);
		
		NodeCommodity * nEntry = client[idx].getNodes();

		while(nEntry!=NULL)
			{
			fprintf(fp, "%s   => %s\n",  
				subPrefixString, 
				nEntry->getNode().getName());
			nEntry = nEntry->getNext();
			}
		}
	fflush(fp);
	delete subPrefixString;
	}
	
