// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	MutualRange.h is the header file that defines the class methods and
//	attributes of the MutualRange class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _MUTUAL_RANGE_H_
#define _MUTUAL_RANGE_H_ 1

#include <RangeTable.h>		// Needed for RangeTable class definition
#include <Brokerage.h>		// Needed for NodeBroker and NodeCommodity
#include <stdio.h>		// Needed for FILE structure and stdio

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class Range;
class JobDeployment;

// *****************************************************************************
// * class MutualRange :
// *	The class is used to create a histogram that identifies when various 
// *	node resources will be simultaneously available.  This histogram will
// *	be used to identify the earliest possible time that a job can executed
// *	based on its specific requirements.
// *****************************************************************************
class MutualRange 
{
private:
	// *********************************************************************
	// * MutualRange * freelist:
	// *	This is the freelist pointer that is used to maintain a pool of
	// *	available MutualRange objects for allocation.
	// *********************************************************************
	static class MutualRange * freeList;

	// *********************************************************************
	// * MutualRange * next:
	// *	This pointer is used to maintain a linked list of MutualRange
	// *	objects.  When the objects are not allocated to a user, this
	// *	next pointer is used to maintain them on the freelist.
	// *********************************************************************
	MutualRange * next;
	
	// *********************************************************************
	// * time_t rangeStart :
	// *	This value indicates the time that this mutual time range
	// *	starts.
	// *********************************************************************
	time_t rangeStart;
	
	// *********************************************************************
	// * time_t rangeDuration :
	// *	Unlike the typical duration flags, this flag indicates the 
	// *	period of time AFTER the rangeStart time that the job can be
	// *	executed and still complete prior to the next scheduled job.
	// *********************************************************************
	time_t rangeDuration;
	
	// *********************************************************************
	// * NodeCommodity * nodes :
	// *	This is an ordered array of NodeCommodity pointers where the
	// *	NodeCommodity objects will be held after allocation. It is
	// *	a staticly allocated array to allow nodes to be directly
	// *	accessed by their node index.
	// *********************************************************************
	NodeCommodity * nodes[SchedStatic::MAXNODES];
	
	// *********************************************************************
	// * MasterBroker broker:
	// *	This is the MasterBroker object that will be used as the 
	// *	central distribution point for all NodeCommodities that are 
	// *	aquired by the MutalRange.  This is a staticly allocated 
	// *	object in order to reduce the time required to perform object 
	// *	creation and memory allocation.
	// *********************************************************************
	MasterBroker broker;
	
	// *********************************************************************
	// * NodeBroker client[SchedStatic::MAXSUBSETS]:
	// *	This is a staticly allocated array of NodeBroker objects that
	// *	represent each of the types that the job requires.  This array
	// *	is allocated to the maximum number of supported subsets, to 
	// *	improve performance and reduce the number of allocations 
	// *	required.
	// *********************************************************************
	NodeBroker client [SchedStatic::MAXSUBSETS];
	
	// *********************************************************************
	// * RangeTable * owner :
	// *	This is the owner of the MutualRange object.  The owner object
	// *	will be accessed in order to notify it that the MutualRange
	// *	object has been satisfied.
	// *********************************************************************
	RangeTable * owner;

	// *********************************************************************
	// * MutualRange::MutualRange :
	// *	This is the default constructor for the object.  Because it
	// *	requires no parameters it should only be called internally
	// *	to repopulate the freelist.
	// *********************************************************************
	MutualRange ( void );

	// *****************************************************************************
	// * MutualRange::addNode :
	// *	This method is called by the MutualRange object when it needs 
	// *	to duplicate its contents in another MutualRange. If a 
	// *	NodeCommodity object that relates to this object exists, then 
	// * 	it will be updated.
	// *****************************************************************************
	void addNode ( NodeCommodity & node );
	
	// *********************************************************************
	// * MutualRange::addNode :
	// *	Once a range has been verified using the add method, this
	// *	method will be called (internally) to add the node and its
	// *	type specification to the MutualRange.
	// *********************************************************************
	void addNode ( ClusterNode & node, size_t type );
	
	// *********************************************************************
	// * MutualRange::split :
	// *	If a range is added that is longer or shorter than the currently
	// * 	defined MutualRange parameters, it will be necessary to split
	// *	the MutualRange into two parts.  This method will cause the
	// *	MutualRange to be split at the specified time.
	// *********************************************************************
	void split ( time_t splitTime );
	
	// *********************************************************************
	// * MutualRange::insert :
	// *	This method is called to insert a new MutualRange object AFTER 
	// *	this MutualRange object.  Note, I'm using an integer for 
	// * 	duration in order to detect a negative duration on systems 
	// * 	where the time_t value is represented by an unsigned integer.
	// *****************************************************************************
	MutualRange * insert ( time_t start, long duration );

	// *********************************************************************
	// * MutualRange::brokerExchange :
	// *	This method is called to attempt to locate available 
	// *	NodeCommodity objects with matching types.  If the number of
	// *	commodity objects in question are located, they will 
	// *	automatically be passed to the client at the specified index 
	// *	and a non-zero value will be returned.  If an available, 
	// *	matching number of commodity objects cannot be located - then 0 
	// *	will be returned indicating that the job cannot be satisfied.
	// *********************************************************************
	int brokerExchange ( size_t idx, size_t cnt = 1 );
	
public:
	// *********************************************************************
	// * MutualRange::operator new :
	// *	This is the new method for the MutualRange object.  It will
	// *	provide a preallocated MutualRange object from the freelist and
	// *	repopulate the freelist when necessary.
	// *********************************************************************
	void * operator new ( size_t size );
	
	// *********************************************************************
	// * MutualRange::operator delete :
	// *	This is the delete method for the MutualRange class.  When  
	// *	a MutualRange object is deleted, this method will return it to 
	// *	the freelist where it can be reallocated to a new user at a
	// *	later time.
	// *********************************************************************
	void operator delete ( void * ptr );
	
	// *********************************************************************
	// * MutualRange::MutualRange :
	// *	This is the standard constructor for the object.  It is called
	// *	with the following parameters.
	// *
	// *	Owner :		This is a reference to the RangeTable that
	// *			owns this MutualRange object.
	// *	StartTime :	This is the earliest time that the range can 
	// *			begin.
	// *	Duration : 	This is the required length of the range.
	// *********************************************************************
	MutualRange ( RangeTable & Owner, time_t StartTime, time_t Duration);
	
	// *********************************************************************
	// * MutualRange::~MutualRange :
	// *	This is the destructor for the MutualRange object.  Because
	// *	this object is always restored to the freelist, this method
	// *	does nothing of consequence.
	// *********************************************************************
	~MutualRange ( void );

	// *********************************************************************
	// * Inline accessor methods that are used to access/modify members.
	// *********************************************************************
	inline time_t        getStartTime ( void ) { return rangeStart; }
	inline time_t        getEndTime   ( void ) { return rangeStart+rangeDuration; }
	inline time_t        getDuration  ( void ) { return rangeDuration; }
	inline MutualRange * getNext      ( void ) { return next; }
	inline void          setNext      ( MutualRange * Next ) { next = Next; }
	
	// *********************************************************************
	// * MutualRange::add :
	// *	This method will walk through the list of MutualRange objects
	// *    and will find the correct location to add a new Range.  The
	// *	method will return a pointer to the first MutualRange object 
	// *	where the Range object was utilized.
	// *********************************************************************
	MutualRange * add ( Range & range, size_t type );
	
	// *********************************************************************
	// * MutualRange::satisfied :
	// *	This method tests to determine if all requirements for the job
	// *	have been satisfied.  It will return a boolean result - TRUE
	// *	means that the MutualRange satifies all requirements, FALSE
	// *	means that the MutualRange does not satisfy all requirements.
	// *********************************************************************
	inline int satisfied ( void ) { return satisfied(owner->getTypeCnt()-1); }	

	// *********************************************************************
	// * MutualRange::satified :
	// *	This is an interim test to ensure that the MutualRange has 
	// *	satisfied the job requirements upto and including the 
	// *	node type specified by typeIdx.  If the job has not satisfied
	// *	its early requirements now, then there is no way that it will
	// *	satisfy them later.
	// *********************************************************************
	int satisfied ( size_t typeIdx );
		
	// *********************************************************************
	// * MutualRange::claim :
	// *	This method is called to transform the MutualRange object into
	// * 	a usable JobDeployment object.  This method will generate the
	// *	empty JobDeployment object and will then populate it by claiming
	// *	the nodes that are currently collected in the subset specific
	// *	NodeBroker clients.
	// *********************************************************************
	JobDeployment * claim ( void );
	
	// *********************************************************************
	// * MutualRange::asciiDump :
	// *	This method is used to generate a diagnostic ascii report of the
	// *	contents of the object. 
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _MUTUAL_RANGE_H */
